import logging
import json
import base64
import os

import jetpack.util
import jetpack.messaging
from jetpack.util.system import service, execute
from jetpack.util import UserError

logger = logging.getLogger(__name__)


def spot_eviction(not_before: str) -> None:
    """ Notifies CycleCloud of a pending spot eviction that will happen `not_before` a certain time.

        `not_before` is expected to be in ISO8601 format
    """
    data = {
        "not_before": not_before
    }
    jetpack.messaging.send_internal_message(data, 'spot_eviction')


def _process_cyclecloud_connection(connection_string=None, connection_file=None):
    ''' Given a connection string or connection filename, generate and write a connection dict to connect to cyclecloud '''
    connection = {}

    if connection_string:
        connection = json.loads(base64.b64decode(connection_string))
    elif connection_file:
        with open(connection_file) as f:
            connection = json.load(f)

    def _override_from_env(v, k):
        ''' Override a key (k) in the connection dict if the env variable (v) exists '''
        if v in os.environ:
            connection[k] = os.environ.get(v)

    _override_from_env('JETPACK_INSTALL_URL', 'url')
    _override_from_env('JETPACK_INSTALL_USERNAME', 'username')
    _override_from_env('JETPACK_INSTALL_PASSWORD', 'password')
    _override_from_env('JETPACK_INSTALL_CLUSTER', 'cluster')
    _override_from_env('JETPACK_INSTALL_CERT', 'cert')
    _override_from_env('JETPACK_INSTALL_CERT_FILE', 'cert_file')
    _override_from_env('JETPACK_INSTALL_PROXY_METHOD', 'proxy_method')

    if connection:
        if not all([v in connection for v in ['url', 'username', 'password', 'cluster']]):
            detail = 'Required: url, username, password, cluster. Found: %s' % ", ".join(list(connection.keys()))
            raise UserError("Incomplete information specified to connect to CycleCloud", detail=detail)

        # If cert_file is specified, read in the 'cert'
        if 'cert_file' in connection:
            with open(connection['cert_file']) as f:
                connection['cert'] = f.read()
            connection.pop('cert_file', None)

        p: str = os.path.join(jetpack.util.jetpack_home(), 'config', 'connection.json')
        with os.fdopen(os.open(p, os.O_WRONLY | os.O_CREAT, 0o640), 'w') as f:
            json.dump(connection, f, indent=4, sort_keys=True)

    return connection


def initialize(connection=None, connection_file=None, cloudinit=False):
    logger.info("Initializing jetpack...")

    if cloudinit and os.name != "nt":
        if not jetpack.util.cloudinit.installed():
            raise UserError("Failed to run Cloud-Init", description="Error running cloud-init: cloud-init is not installed")
        if not jetpack.util.cloudinit.enabled():
            raise UserError("Failed to run Cloud-Init", description="Error running cloud-init: cloud-init is installed but disabled")

        logger.info("Waiting for Cloud-Init to complete")
        errors = jetpack.util.cloudinit.wait()
        if len(errors) > 0:
            logger.error("Cloud-init completed with errors: %s" % "\n".join(errors))
            raise UserError("Failed to run Cloud-Init", description="Detected error(s) running cloud-init", detail="\n".join(errors))

        logger.info("Cloud-init completed successfully")

    # Ensure that a valid connection was sent
    if not _process_cyclecloud_connection(connection, connection_file):
        raise UserError("No connection information was provided to communicate with CycleCloud")

    # Validate the CycleCloud connection
    jetpack.util.test_cyclecloud_connection()
    logger.info("Jetpack connection to CycleCloud verified!")


def start_service():
    # Enable and start the jetpack services
    if jetpack.util.is_linux():
        service('jetpackd', enable=True)
        service('healthcheck', enable=True)
        service('jetpackd', start=True)
        service('healthcheck', start=True)
    else:
        service('jetpackd', start=True)
        execute('schtasks /Create /TN HealthCheck /F /SC minute /MO 5 /RU SYSTEM /RL HIGHEST /TR "C:\\cycle\\jetpack\\system\\bin\\healthcheck.cmd --logfile=C:\\cycle\\jetpack\\logs\\healthcheck.log"', shell=True, msg="Created scheduled task 'HealthCheck'")
