import os
import logging
import json
from jetpack import util
from typing import Any, Dict

logger = logging.getLogger()

UNKNOWN_ERROR = "An unknown error occurred while processing the configuration data file."


class ConfigError(Exception):
    pass


def get(value: str = None, default: Any = None, configuration_file: str = None) -> Any:
    try:
        current_val = _get_data(configuration_file)
        if value is None:
            return current_val
        properties = value.split('.')
        for prop in properties:
            current_val = current_val[str(prop)]
        return current_val
    except KeyError:
        if default is not None:
            return default
        else:
            raise ConfigError("The value '%s' was not found in the node configuration." % value)
    except Exception:
        raise ConfigError(UNKNOWN_ERROR)


def available() -> bool:
    """ Checks to see if the settings are available to be queried yet """
    return os.path.isfile(_config_file())


def update(config: Dict[Any, Any]) -> Dict[Any, Any]:
    """ Update the client-config.json file used by chef (which creates node.json)

    NOTE: In the future client-config.json will be $JEPTACK_HOME/config/config.json and will take the place 
    of node.json
    NOTE: Updating client-config json DOES NOT update node.json, that requires a reconverge (sadness)
    meaning an update followed by a get will not return the right thing (yet)
    """
    filename = _chef_config_file()
    current_config = {}
    if os.path.isfile(filename):
        with open(filename, 'r') as f:
            current_config = json.load(f)

    util.deep_merge(config, current_config)

    with open(filename, 'w') as f:
        f.write(json.dumps(current_config, indent=4, sort_keys=True))

    return config


def _chef_config_file() -> str:
    """ Gets the client-config.json file
    NOTE: This will be moved to $JETPACK_HOME/config/config.json in the future
     """
    return os.path.join(util.jetpack_home(), "system", "chef", "client-config.json")


def _config_file() -> str:
    """ Get the node.json configuration based on the OS """
    return os.path.join(util.jetpack_home(), "config", "node.json")


def _get_data(config_file: str = None) -> Dict[Any, Any]:
    """ Return a dictionary of data from a config file """
    import json
    if config_file is None:
        config_file = _config_file()

    config_file = os.path.expanduser(config_file)

    try:
        # node.json will be utf8 encoded since the default encoding of ruby 
        # (which produces the file) is utf8
        if os.path.isfile(config_file):
            with open(config_file, encoding='utf-8') as f:
                data = json.loads(f.read())
                return data
        else:
            import jetpack.converge.userdata
            logger.debug("%s not found, using config from userdata instead", config_file)
            return jetpack.converge.userdata.get_config()

    except IOError as e:
        raise ConfigError(e.strerror)
    except ValueError as e:
        raise ConfigError(str(e))
    except Exception:
        logger.exception("Unhandled exception querying config data")
        raise ConfigError(UNKNOWN_ERROR)
