import subprocess
from jetpack import util
import os
import json
import distro  # type: ignore
import datetime
from jetpack.converge import userdata
from jetpack import messaging
import jetpack.util
import logging
from jetpack import config as jetpack_config
import jetpack.cluster_init
from jetpack.admin import configure
import jetpack.settings
import jetpack.lockers
import socket

import jetpack.status

logger = logging.getLogger()

CONFIGURATION_STATUS_FILE = os.path.join(jetpack.util.jetpack_home(), 'system', 'run', 'configuration-status.json')

class ConvergeError(Exception):
    def __init__(self, *args, retryable_error: bool = True):
        super().__init__(*args)
        self.retryable_error = retryable_error

class EncryptionError(Exception):
    pass

class InstanceNotReady(Exception):
    pass

def _setup_proxy():
    ''' Setup the proxy environment if the settings were passed in via [[[configuration]]]'''
    # TODO: This needs a unittest
    if jetpack.util.is_linux():
        proxy_url = jetpack_config.get('http_proxy.url', '')
        if proxy_url:
            # Write a profile script with the proxy settings
            proxy_profile_file = '/etc/profile.d/proxy.sh'
            if not os.path.isfile(proxy_profile_file):
                with open(proxy_profile_file, 'w') as f:
                    f.write("# Written by CycleCloud\n")
                    f.write("export http_proxy=%s\n" % proxy_url)
                    f.write("export https_proxy=%s\n" % proxy_url)
                    f.write("export no_proxy=169.254.169.254  # special rule exempting metadata URL from proxy\n")
                os.chmod(proxy_profile_file, 0o644)
            
            # Update the current environment with the proxy settings since the above profile won't be loaded
            os.environ['http_proxy'] = proxy_url
            os.environ['https_proxy'] = proxy_url
            os.environ['no_proxy'] = '169.254.169.254'


def _sync_chef_repos(bootstrap_path, log_path):
    # Handle CycleCloud Project specs by using the API
    # Iterates over every spec (if any) assigned to the node, and downloads, untars, and moves
    # appropriate files into $JETPACK_HOME/system/chef/chef-repo/cookbooks|data_bags|roles in order
    # overwriting full directories as needed.
    jetpack.cluster_init.process_chef()


def _run_chef(cmd):
    p = subprocess.Popen(cmd, shell=True)
    p.communicate()
    # Chef returns a special exit code if a reboot has been requested
    # https://github.com/chef/chef-rfc/blob/master/rfc062-exit-status.md#exit-codes-in-use
    if util.is_windows() and p.returncode == 35:
        return "rebooting"
    elif p.returncode != 0:
        raise ConvergeError("A problem occurred while running Chef, check chef-client.log for details")
    else:
        return 'success'


def _run_chef_solo(log_path, flags):
    if util.is_linux():
        cmd = 'chef-solo -c /opt/cycle/jetpack/system/chef/solo.rb -L %s/chef-client.log %s' % (log_path, flags)
    else:
        cmd = 'chef-solo.bat -c C:\\cycle\\jetpack\\system\\chef\\solo.rb -L %s\\chef-client.log %s' % (log_path, flags)

    return _run_chef(cmd)


def _run_jetpack_test(previously_run=False):
    try:
        if previously_run:
            logger.info("Integration tests already run, skipping")
            report = jetpack.cluster_init.repost_test_report()
        else:
            report = jetpack.cluster_init.run_tests()

        if report.failed:
            raise jetpack.cluster_init.JetpackTestException(report.error_message())

        return 'success'

    except Exception as e:
        raise ConvergeError("A problem occurred with the cluster-init integration test, message: %s" % str(e), retryable_error=False)


def _send_installation_status(status):
    message = _read_timestamps()
    message['status'] = status

    # New versions of jetpack don't use the installation message to track the status of a node, instead the `jetpack.status` module
    # handles sending back a larger overall view of the node status. Tell the installation status listener that we can expect
    # jetpack.status to be used for the status instead of the installation listener. Older versions of jetpack WILL NOT send this key
    # allowing CycleCloud to retain the original behavior for legacy clusters.
    message['update_status'] = False
    try:
        message['hostname'] = socket.gethostname()
        message['fqdn'] = socket.getfqdn()
    except Exception as e:
        logger.error("Could not get hostname information: %s" % str(e))

    try:
        messaging.send_internal_message(message, "installation")

    except Exception as e:
        logger.error("Could not send status message: %s" % str(e))
        raise e

def _update_runlist(config):
    for spec in jetpack.cluster_init.get_specs():
        if spec.run_list is not None:
            config['run_list'] += spec.run_list

    # Add cyclecloud and cluster_init to runlist if they aren't already there
    config['run_list'] = config.get("run_list", [])
    if "recipe[cyclecloud]" not in config['run_list']:
        config['run_list'].insert(0, "recipe[cyclecloud]")
    if "recipe[cluster_init]" not in config['run_list']:
        config['run_list'].append("recipe[cluster_init]")

    return config

def _coerce_platform_family(config):
    """ Coerce the platform_family for newer distros like Alma and Rocky Linux since our version of ohai
    does not support these distributions """
    if distro.id().lower() in ['almalinux', 'rocky']:
        config['platform_family'] = 'rhel'
    if distro.id().lower() in ['sle-hpc', 'sle_hpc']:
        config['platform_family'] = 'suse'
    return config


def _read_timestamps():
    # Reads the timestamp file from disk and returns it as a dictionary
    timestamp_file = os.path.join(util.jetpack_home(), 'config', 'timestamps.json')
    if os.path.isfile(timestamp_file):
        with open(timestamp_file, 'r') as f:
            return json.load(f)
    else:
        return {}


def _write_configuration():
    try:
        config = configure.execute()
        return config
    except configure.JetpackConfigureError as e:
        raise ConvergeError(str(e))


def execute(debug=False, no_sync=False, mode=None):
    """ Shells out to the converge script that is already on the box """

    if util.is_windows():
        bootstrap_path = 'c:\\cycle\\jetpack\\system\\bootstrap'
        log_path = 'c:\\cycle\\jetpack\\logs'
        pid_file = 'c:\\cycle\\jetpack\\system\\run\\converge.pid'
    else:
        bootstrap_path = "/opt/cycle/jetpack/system/bootstrap"
        log_path = '/opt/cycle/jetpack/logs'
        pid_file = '/opt/cycle/jetpack/system/run/converge.pid'

    # This should only run once at a time
    if util.already_running(pid_file):
        raise ConvergeError("Converge process already running...")

    ###########################################
    # Track the time we first started #
    ###########################################
    timestamps = jetpack.util.write_timestamps({
        "converge_initial_start_timestamp": datetime.datetime.now(),
        "boot_timestamp": util.get_boottime(),
    })
    # # Grab the initial timestamp from whatever was just read, it's the initial converge if the current time 
    # # is the one that was just written
    start_time = timestamps.get('converge_initial_start_timestamp')

    logger.info("Beginning converge")

    # Clear out our configuration-status.json file each converge
    jetpack.util.clear_configuration_status()

    # NOTE: This is done with multiple if linux else windows checks because the
    # subprocess calls will be replaced with OS independent python code in the future

    ##############################################
    # Cache userdata and write client config #
    ##############################################
    config = userdata.get_config()

    # HACK: coerce the platform family to RHEL for almalinux and rocky linux since our version of ohai does not support them
    config = _coerce_platform_family(config)

    # have to write out the config.ini file first because to set status we connect to CC
    # and to update the runlist we have to download the manifest which requires keys in the config
    # NOTE: there may be a better way to do this now that connection string is passed on jetpack installation
    _write_configuration()

    jetpack_config.update(config)

    status = "failure"
    retryable = True
    try:

        #############################
        # Send booted status signal #
        #############################

        # Send the initial software configuration status
        # if it hasn't yet been sent
        if mode == 'install' and 'configuration_status_timestamp' not in timestamps:
            jetpack.status.set(jetpack.status.JetpackResource.SoftwareConfiguration,
                            status=jetpack.status.JetpackStatus.NoStatus,
                            message="Configuring software",
                            description="Configuring software on VM",
                            active=True,
                            start_time=start_time)
            jetpack.util.write_timestamps({
                'configuration_status_timestamp': datetime.datetime.now()
            })

        _send_installation_status("started")


        ##########################################
        # Configuration                          #
        ##########################################
        config = _update_runlist(config)
        config = jetpack.util.get_current_config(config)
        
        node_config = config["cyclecloud"]["node"]
        if not node_config["ready"]:
            raise InstanceNotReady()

        jetpack_config.update(config)

        # Write out the settings.json and lockers.json files
        # NOTE: For now we just get that data from legacy userata, but we will pull it live from CC in the future
        # NOTE: Custom data is a json string embedded within the legacy XML
        jetpack.settings.update(userdata.get_settings())
        jetpack.lockers.update(userdata.get_lockers())
        
        # (Legacy/Deprecated) 
        # Setup the proxy configuration
        # - This is supported for legacy compatability only
        # - Newer clusters should probably use cloud-init for proxy configuration
        _setup_proxy()

        # Disable SELinux if specified by the user
        # NOTE: cyclecloud.selinux.policy is a carry over attribute form CycleCloud 7.x, it is no longer set in Chef by default
        # but a user can pass this in to change the mode on their VM at runtime
        if jetpack_config.get('cyclecloud.selinux.policy', ""):
            jetpack.util.selinux_mode(jetpack_config.get('cyclecloud.selinux.policy', "").lower())

        # Check to see if we should continue based on the encryption status of the disk jetpack is on.
        if jetpack_config.get(value="cyclecloud.await_encryption", default=False):
            if util.is_windows():
                drive = os.path.splitdrive(util.jetpack_home())[0]
                try:
                    output = subprocess.check_output(['manage-bde', '-status', drive])
                    if "Fully Decrypted" in output:
                        raise EncryptionError("The disk reported as being fully decrypted")
                except (subprocess.CalledProcessError, EncryptionError):
                    message = "Expected %s drive to be encrypted; external encryption process not started yet" % drive
                    logger.error(message)
                    raise ConvergeError(message)
            else:
                logger.warning("Expected drive to be encrypted, but cannot check linux encryption status; continuing...")

        # Legacy mode happens when we don't have a "cyclecloud" project defined
        # in those cases we will use the shared chef repos
        spec_count = len(jetpack.cluster_init.get_specs(project_name="cyclecloud"))
        legacy_mode = spec_count == 0
        jetpack_config.update({'cyclecloud': {'legacy_common_chef_repo': legacy_mode}})

        #############################
        # Configure jetpack CLI #
        #############################
        _write_configuration()

        ############################################
        # Sync Chef repos to disk and run Chef #
        ############################################
        flags = ""
        if debug:
            flags = "-l debug"

        if no_sync is False:
            _sync_chef_repos(bootstrap_path, log_path)

        status = _run_chef_solo(log_path, flags)

        # if the chef converge has triggered a reboot, skip remaining steps and
        # go to the finally block
        if status == 'rebooting':
            return
        
        ##############################################
        # Execute cluster-init integration tests #
        ##############################################
        test_mode = jetpack_config.get(value="cyclecloud.test_mode", default=False)

        if test_mode:
            status = "failure"
            previously_run = os.path.exists(os.path.join(log_path, 'test-results.json'))
            if not previously_run:
                _send_installation_status("running_tests")

            status = _run_jetpack_test(previously_run)
    except Exception as e:
        jetpack.util.write_configuration_exception(e)

        # Attempt to rebuild the RPM database on any error
        # If the db is rebuilt, change the status from 'failure' to 'warning' and raise an appropriate
        # error that the db is being repaired
        if jetpack.util.rebuild_rpmdb():
            e = jetpack.util.UserError(message='Rebuilding corrupt RPM database', 
                                       description='Detected corrupt RPM database, rebuilding...')

            # Write the updated error to disk overwriting the previous one detailing that the DB was corrupt
            # since we just fixed it
            util.write_configuration_exception(e, force=True)
            status = 'warning'

        if isinstance(e, ConvergeError):
            retryable = e.retryable_error

        logger.error("Converge failed: %s" % str(e))

        raise e
    finally:
        # If we are in 'install' mode, the status is active until we succeed
        end_time = datetime.datetime.now().isoformat()

        jetpack_status = {}

        # Report the overall status via the jetpack.status API as well
        if status == 'success':
            # Write the timestamp for the successful converge
            timestamps = jetpack.util.write_timestamps({"converge_initial_finish_timestamp": datetime.datetime.now()})
            end_time = timestamps.get('converge_initial_finish_timestamp')

            jetpack_status = {
                'name': jetpack.status.JetpackResource.SoftwareConfiguration,
                'status': jetpack.status.JetpackStatus.Ok,
                'message': "Completed software configuration",
                'description': "Completed software configuration",
                'active': False,
                'start_time': start_time,
                'end_time': end_time
            }
    
        elif status in ['warning', 'failure']:
            # On a converge failure, read the converge-status from 
            # disk which contains information about what exactly failed
            # during the converge. Send this information back to CC (and log it locally)
            config_status = jetpack.util.read_configuration_status()

            jetpack_status = {
                'name': jetpack.status.JetpackResource.SoftwareConfiguration,
                'status': jetpack.status.JetpackStatus.Error,
                'message': config_status.get('message') or "Error configuring software",
                'description': config_status.get('description'),
                'detail': config_status.get('detail'),
                'recommendation': config_status.get('recommendation'),
                'help_link': config_status.get('help_link'),
                'active': True,
                'start_time': start_time,
            }

            if status == 'warning':
                jetpack_status['status'] = jetpack.status.JetpackStatus.Warning

            if not retryable:
                # Write the timestamp for the failed converge
                timestamps = jetpack.util.write_timestamps({"converge_initial_finish_timestamp": datetime.datetime.now()})
                end_time = timestamps.get('converge_initial_finish_timestamp')
                jetpack_status['active'] = False

            logger.error("Software configuration failed: %s\n" % config_status.get('description'))
            if config_status.get('recommendation'):
                logger.error("Recommendation: %s" % config_status.get('recommendation'))

        else: # at very least the 'rebooting' status is handled here, not sure it should be...
            jetpack_status = {
                'name': jetpack.status.JetpackResource.SoftwareConfiguration,
                'status': jetpack.status.JetpackStatus.Error,
                'message': "Error configuring software",
                'description': "Unknown configuration status returned: '%s'" % status,
                'active': True,
                'start_time': start_time
            }

        # Report the node status using the old API for backwards compatibility
        # NOTE: We don't report reboots as a converge status    
        if status != 'rebooting':
            _send_installation_status(status)

        # Send a status update if we're in install mode
        if mode == 'install':
            jetpack.status.set(**jetpack_status)

        # Clean up the converge.pid since the process has ended
        if os.path.isfile(pid_file):
            os.remove(pid_file)

    if status == 'success':
        jetpack.status.update(jetpack.status.JetpackResource.SoftwareConfiguration, jetpack.status.JetpackStatus.Ok, "Cluster converged successfully")
        logger.info("Converge succeeded")
        return True

    return False
