import json
import logging
import os
import time
from typing import Dict, Any, Optional
import urllib.parse

import jetpack.util
import jetpack.converge.userdata

UNKNOWN_ERROR = "An unknown error occurred while processing the lockers data file."

logger = logging.getLogger(__name__)


class LockerError(Exception):
    pass


def update(data: Dict[Any, Any]) -> None:
    """ Updates the lockers.json file on disk with the new locker data """
    with open(_lockers_file(), 'w') as f:
        f.write(json.dumps(data, indent=4, sort_keys=True))


class Locker():
    def __init__(self, name: str, data: Dict[Any, Any]):
        """ Loads a locker from data passed in via customdata """
        self._creds: Dict[Any, Any] = data.get('creds', {})
        self.name: str = name
        self.region: str = str(data.get('region'))
        self.account_name: str = str(data.get('account_name'))
        self.location: str = str(data.get('location'))

        self._base_url: Optional[str] = None
    
    def url(self, path: str, include_sas_token: bool = False) -> str:
        """ Gets the URL for a path within the locker, optionally included a sas token """
        if path[0] == "/":
            raise ValueError("Unable to generate url for locker with a path that starts with a /: %s" % path)

        full_url = "%s/%s" % (self.base_url, path)
        
        if include_sas_token and self.client_id is None:
            # the SAS token we have might have almost expired - allow some extra time for the download
            # downloading the file is faster than full converge, so 5 minutes should be enough
            if time.time() > self._creds.get('sas_token_valid_until', 0) - 300:
                logger.info("SAS token for locker %s has expired, a new one was requested" % self.name)
                updated_lockers = jetpack.converge.userdata.get_lockers(refresh_cache=True)
                updated_locker_data = updated_lockers.get(self.name, {})
                self._creds = updated_locker_data.get('creds', {})
                # update the locker data
                updated_lockers.update({self.name: updated_locker_data})
                update(updated_lockers)
            full_url += self._creds.get('sas_token', '')

        return full_url

    def __str__(self):
        return "%s (%s)" % (self.name, self.location)

    @property
    def base_url(self) -> str:
        """ Gets the base URL for the Locker"""
        if self._base_url is None:
            r = urllib.parse.urlparse(self.location)
            base = "https://%s.blob%s" % (r.netloc, self._creds.get('environment', {}).get('storage_suffix', ''))
            self._base_url = urllib.parse.urljoin(base, r.path).rstrip("/")
        return self._base_url

    @property
    def creds(self):
        return self._creds
    
    @property
    def client_id(self) -> str:
        return self._creds.get('client_id')

def _get_data() -> Dict[Any, Any]:
    """ Fetches the lockers.json from disk or userdata """
    if not os.path.isfile(_lockers_file()):
        return jetpack.converge.userdata.get_lockers()
    else:
        try:
            with open(_lockers_file()) as f:
                return json.loads(f.read())
        except IOError as e:
            raise LockerError(e.strerror)


def get(locker_name: str) -> Locker:
    """ Gets a locker by name """
    data = _get_data()
    if locker_name in data:
        return Locker(name=locker_name, data=data.get(locker_name, None))
    else:
        raise LockerError("Unable to find a locker named '%s' in locker config" % locker_name)


def _lockers_file() -> str:
    """ Gets the path to the lockers.json file """
    return os.path.join(jetpack.util.jetpack_home(), 'config', 'lockers.json')
