from jetpack import util
import time
import json


class SendError(Exception):
    pass


class LogError(Exception):
    pass


def log(message, level, priority=None):
    if level not in ["info", "warn", "error"]:
        raise LogError("Invalid level: %s" % level)

    priority = priority or _get_priority(level)

    message_data = {
        "level": level,
        "message": message,
        "priority": priority
    }
    
    send_internal_message(message_data, "log")


def send_internal_message(message_data, message_type, config=None):
    '''
    Sends a system message to CycleCloud
    parameters:
      message_data - this is a python dictionary
      message_type - type of message, examples are test, log, installation
    '''
    if not isinstance(message_data, dict):
        raise SendError("message_data parameter must be a dictionary")
    try:
        if config is None:
            config = util.parse_config(None)
        identity = config["identity"]
        cluster_session_id = identity.get('cluster_session_id')
        cluster_name = identity["cluster_name"]
        instance_id = identity["instance_id"]
        cycle_server_config = config['cycle_server']
    except KeyError as e:
        raise SendError("Unable to find '%s' in config" % str(e))

    message_obj = {
        "cluster_name": cluster_name,
        "instance_id": instance_id,
        "timestamp": util.iso_8601_timestamp(),
        "cluster_session_id": cluster_session_id,
        "type": message_type,
        "data": message_data
    }

    def func():
        r = _post_message(message_obj)
        if r.status != 202:
            raise Exception("Failed to send message: %d" % r.status)

    return _retry_func(func)


def _post_message(message_obj):
    return util.query_cyclecloud("/clusterlink/messages", body=json.dumps({"messages": [message_obj]}), method="POST")


def _retry_func(func):
    wait_length = 5

    while True:
        try:
            return func()
        except Exception as e:
            # retry 5 times, waiting 5, 10, 20, then 40 seconds
            if wait_length < 41:
                time.sleep(wait_length)
                wait_length *= 2
            else:
                raise e

def _get_priority(level):
    if level == 'info':
        log_priority = 'medium'
    elif level == 'warn':
        log_priority = 'medium'
    elif level == 'error':
        log_priority = 'high'
    else:
        raise LogError("Invalid log level")
    return log_priority
