import datetime
import logging
import threading
import time
import urllib.request
import json
import os
from subprocess import Popen
from typing import Optional, Any, Dict, List, Set

from jetpack.util import JETPACK_USER_AGENT
import jetpack.admin
import jetpack.config
import jetpack.messaging
from jetpack import shutdown_actions

logger = logging.getLogger(__name__)

SCHEDULED_EVENT_SCRIPT_DIR = os.path.join(jetpack.util.jetpack_home(), 'scripts')

class TransientEventMonitorException(Exception):
    """ Exception that is likely a transient error and should just be retried """
    pass


class EventMonitorException(Exception):
    pass

class EventMonitorUserDefinedScriptException(Exception):
    pass


class ScheduledEventMonitor(object):
    def __init__(self, resource_name: Optional[str] = None, interval: int = 1):
        self.resource_name: str = resource_name or jetpack.util.get_resource_name()
        self.interval: int = interval
        self.api_version: str = '2019-08-01'
        self.endpoint: str = 'http://169.254.169.254/metadata/scheduledevents'

        # The list of eventIDs that has previously been processed
        self.handled_events: Set[str] = set()

    def _get_script_name(self, event_type: str) -> str:
        script_name = "on" + event_type + '.sh'
        if jetpack.util.is_windows():
            script_name = "on" + event_type + '.bat'  
        return script_name

    def execute_userdefined_script(self, event_type: str) -> int:
        """ Attempt to run a user defined script and return a bool representing
        success or failure of the script """

        script_name = self._get_script_name(event_type)
        script_path = os.path.join(SCHEDULED_EVENT_SCRIPT_DIR, script_name)

        logger.info("Executing user defined %s notification script" % event_type)
        try:
            with open(os.path.join(SCHEDULED_EVENT_SCRIPT_DIR, "%s.out" % event_type), 'w') as stdout:
                with open(os.path.join(SCHEDULED_EVENT_SCRIPT_DIR, "%s.err" % event_type), 'w') as stderr:
                    if jetpack.util.is_windows():
                        p = Popen(script_path, stdout=stdout, stderr=stderr, cwd=SCHEDULED_EVENT_SCRIPT_DIR, shell=True)
                    else:
                        p = Popen([script_path], stdout=stdout, stderr=stderr, cwd=SCHEDULED_EVENT_SCRIPT_DIR)
                    p.wait()

                    if p.returncode == 0:
                        logger.info("Successfully ran %s notification script" % script_name)
                    else:
                        logger.error("Error executing %s notification script - unexpected exit status %s" % (script_name, p.returncode))

                    return p.returncode
        except Exception:
            msg = "An error occurred while trying to execute %s notification script" % script_name
            logger.exception(msg)
            raise EventMonitorUserDefinedScriptException(msg)

    def start(self):
        thread = threading.Thread(target=self.run, args=())
        thread.daemon = True
        thread.start()
        logger.info("Scheduled event monitor thread started")

    def run(self):
        while True:
            try:
                self.check_for_events()
            except Exception:
                logger.exception("Unhandled exception in scheduled event monitor")

            time.sleep(self.interval)

    def _validate_event(self, event: Dict[str, Any]):
        """ Validate an event to ensure it has the expected structure """
        if not isinstance(event, dict):
            raise EventMonitorException("Invalid event: expected dict but got %s" % type(event))
        for key in ['EventId', 'EventType', 'Resources']:
            if key not in event:
                raise EventMonitorException("Invalid event: missing required key %s: %s" % (key, event))
            if not event[key]:
                raise EventMonitorException("Invalid event: key %s is None or empty: %s" % (key, event))
            
    def check_for_events(self):
        events = []

        try:
            events = self.get_scheduled_events()
        except EventMonitorException as e:
            logger.error(str(e))
            return
        except TransientEventMonitorException as e:
            logger.error(str(e))
            return

        acknowledged_events: Set[str] = set()
        event_ids: Set[str] = set()

        received_shutdown_event = False
        for event in events:    

            try:
                self._validate_event(event)
            except EventMonitorException as e:
                logger.error("Skipping Invalid ScheduledEvent: %s" % str(e))
                continue

            # Only run shutdown actions for events that are relevant to this resource
            if self.resource_name in event['Resources']:                
                received_shutdown_event = received_shutdown_event or event['EventType'].lower() in ['terminate', 'preempt']

            event_ids.add(event['EventId'])
            if event['EventId'] in self.handled_events:
                continue
            if event['EventType'].lower() == 'preempt':
                if self.resource_name in event['Resources']:
                    logger.info("Spot preemption event detected")
                    not_before = datetime.datetime.strptime(event['NotBefore'], '%a, %d %b %Y %H:%M:%S GMT').isoformat() + "Z"

                    # Alert CycleCloud of the impending preemption
                    jetpack.admin.spot_eviction(not_before)
                    logger.info("Spot preemption event reported to CycleServer occurring at %s" % not_before)

                    # Try to run the user-defined script prior to preemption
                    if os.path.isfile(os.path.join(SCHEDULED_EVENT_SCRIPT_DIR, self._get_script_name(event['EventType']))):
                        try:
                            if self.execute_userdefined_script('Preempt') == 0:
                                self.handled_events.add(event['EventId'])
                        except EventMonitorUserDefinedScriptException:
                            pass  # try again next time, don't let exceptions kill the monitor
                    else:
                        self.handled_events.add(event['EventId'])
            else:
                if os.path.isfile(os.path.join(SCHEDULED_EVENT_SCRIPT_DIR, self._get_script_name(event['EventType']))):
                    if self.resource_name in event['Resources']:
                        returncode = None
                        try:
                            returncode = self.execute_userdefined_script(event['EventType'])
                        except EventMonitorUserDefinedScriptException as e:
                            returncode = f"'error: {str(e)}'"

                        status = "not acknowledging"
                        if returncode == 0:
                            status = "acknowledging"
                            acknowledged_events.add(event['EventId'])
                        
                        name = jetpack.config.get('name', self.resource_name)
                        msg = "%s event detected for %s: %s exited with exit code %s, %s event" % (event['EventType'], name, self._get_script_name(event['EventType']), returncode, status)
                        logger.info(msg)
                        jetpack.messaging.log(msg, level='info')
                    else:
                        # If the event is for another resource, assume another instance of Jetpack will handle the event
                        pass
                else:
                    status = "not acknowledging"
                    if event.get('EventSource', '').lower() == 'user':
                        # Automatically acknowledge any user defined events
                        # If we do not, Azure puts a automatic 15 minute delay on events like
                        # user initiated reboots, if we don't auto-acknowledge them users will have
                        # to wait far longer than they should.
                        status = "acknowledging"
                        logger.info("Auto-acknowledging EventID %s for resources %s (%s)" % (event['EventId'], event['Resources'], event['EventType']))
                        acknowledged_events.add(event['EventId'])
                    else:
                        # Otherwise - system defined events can just be timed out as normal
                        # this allows unhandled events like scheduled maintenance to timeout on their usual cadence if the user hasn't
                        # specifically acknowledged the event.
                        logger.info("Ignoring system defined event %s for resources %s (%s)" % (event['EventId'], event['Resources'], event['EventType']))

                        self.handled_events.add(event['EventId'])

                     # If the current resource is part of the event, log it to CycleCloud for information purposes but otherwise take no action
                    if self.resource_name in event['Resources']:
                        name = jetpack.config.get('name', self.resource_name)
                        if event['EventType'].lower() in ['terminate', 'preempt']:
                            # Acknowledge shutdown events for this node even if there are no registered handlers so that termination is quick
                            status = "acknowledging"
                            msg = "%s %s initiated event detected for %s scheduled for %s: no local handlers found, %s event" % (event['EventType'], event.get('EventSource', 'unknown').lower(), name, event.get('NotBefore', 'n/a'), status)
                            acknowledged_events.add(event['EventId'])
                            logger.warn(msg)
                            # log at info level since we're going to acknowledge Termination events...   
                            # This may still be unexpected if the user intended to register an onTerminate, but it's too spammy
                            jetpack.messaging.log(msg, level='info')

                        else:
                            jetpack.messaging.log("%s %s initiated event detected for %s scheduled for %s: no local handlers found, %s event" % (event['EventType'], event.get('EventSource', 'unknown').lower(), name, event.get('NotBefore', 'n/a'), status), level='info')

        if received_shutdown_event:
            # Execute shutdown actions ahead of termination/preemption   
            shutdown_actions.handle_shutdown()

        if acknowledged_events:
            try:
                self.start_scheduled_events(list(acknowledged_events))
                self.handled_events.update(acknowledged_events)
            except EventMonitorException:
                logger.exception("Failed to acknowledge scheduled events: %s" % acknowledged_events)

        # Remove any handled events that no longer show up in the event list since they have passed and no longer are even reported
        # by the service
        self.handled_events &= event_ids

    def _query_scheduled_events_endpoint(self, data=None):
        url = "%s?api-version=%s" % (self.endpoint, self.api_version)

        if data:
            data = data.encode()

        request = urllib.request.Request(url, data, {"Metadata": "true", 'User-Agent': JETPACK_USER_AGENT})
        try:
            response = urllib.request.urlopen(request)
            return response.read()
        except urllib.error.HTTPError as e:
            raise EventMonitorException("Expected status 200 but got %s interacting with scheduled events endpoint" % e.code)
        except urllib.error.URLError as e:
            raise EventMonitorException("Unable to interact with scheduled events endpoint: %s" % e.reason)

    def get_scheduled_events(self) -> Any:
        """ Queries the scheduled events endpoint and returns the JSON representation of events """
        
        data = None
        try:
            data = self._query_scheduled_events_endpoint()
            if len(data) == 0:
                raise TransientEventMonitorException("Empty response from scheduled events likely due to rate limiting, try again")
            event_data = json.loads(data)
        except json.JSONDecodeError as e:
            raise EventMonitorException("Unable to parse scheduled events from server: %s" % str(data))
        
        try:
            events = event_data['Events']
        except KeyError:
            raise EventMonitorException("Invalid event: 'Events' key missing in scheduled events from server: %s" % str(data))
        
        # Return the embedded list of events, not the wrapped response
        return events

    def start_scheduled_events(self, event_ids: List[str]):
        """ Acknowledges a set of EventIDs so that Azure can take action before their NotBefore time passes """
        request: Dict[str, List[Dict[str, str]]] = {
            "StartRequests": []
        }
        for event_id in event_ids:
            request['StartRequests'].append({"EventId": event_id})

        self._query_scheduled_events_endpoint(json.dumps(request, sort_keys=True))
