
import logging
import jetpack.config
import jetpack.util
from jetpack import shutdown_actions
import subprocess
import urllib.request, urllib.parse, urllib.error
import time
import sys
import os
import json
import http.client


logger = logging.getLogger()

class ShutdownError(Exception):
    pass


def execute(idle=False, unhealthy=False, deallocate=False, terminate=False, max_attempts=5, wait=5, config=None):
    """ Contacts CycleCloud to request a shutdown, forces shutdown as needed 

    max_attempts is the number of times to make a request to CC before failing
    wait is the time between each attempt (doubled every failure)
    """

    # Write a shutdown pid, and validate
    pid_file = os.path.join(jetpack.util.jetpack_home(), 'system', 'run', 'shutdown.pid')    
    if jetpack.util.already_running(pid_file):
        raise ShutdownError("Shutdown process already running...")

    if idle and unhealthy:
        raise ShutdownError("Nodes cannot be both 'idle' and 'unhealthy' when requesting shutdown")

    reason = "none"
    if idle:
        reason = "idle"
    if unhealthy:
        reason = "unhealthy"

    action = None
    if deallocate and terminate:
        raise ShutdownError("Both 'deallocate' and 'terminate' cannot be specified at the same time")
    elif deallocate:
        action = "deallocate"
    elif terminate:
        action = "terminate"

    # Ignore max attempts in azure
    max_attempts = None

    # Loop forever trying to shutdown until we hit a max tries
    attempts = 0
    try:
        while True:
            attempts += 1

            successful = True
            try:
                if action:
                    request_action(action)
                else:
                    request_shutdown(reason, config)
            except Exception as e:
                successful = False
                logger.error(str(e))
                print("%s (attempt %s/%s)" % (str(e), attempts, 'infinite'))

            # Note: Azure doesn't actually delete VMs on shutdown, so we will just keep requesting shutdown
            # until the node goes away

            # Wait between requests (max: 90 seconds)
            wait = min(wait * (attempts), 90)
            time.sleep(wait)
    finally:
        if os.path.isfile(pid_file):
            os.remove(pid_file)


def request_action(action: str) -> None:
    if action not in ['shutdown', 'deallocate', 'terminate']:
        raise ShutdownError("Unknown action: %s" % action)

    configuration = jetpack.util.parse_config()

    cluster_name = configuration['identity']['cluster_name']
    node_id = jetpack.config.get('cyclecloud.node.id')
    if node_id is None:
        raise ShutdownError("Unable to determine node id")

    url = urllib.parse.quote("/clusters/%s/nodes/%s" % (cluster_name, action))
    body = json.dumps({'ids': [node_id]})
    try:
        r = jetpack.util.query_cyclecloud(url, body=body, method="POST")
        if r.status == 202:
            # 202 means the operation succeeded, but we should look to see if we got an error back for the specific node
            # we are taking an action on
            data = json.loads(r.read())
            for node in data.get('nodes', []):
                if node.get('id') == node_id and node.get('status') == "Error":
                    raise ShutdownError("Unable to %s node %s: %s" % (action, node_id, node.get('error')))
        else:
            raise ShutdownError("Unexpected response trying to %s node %s: %s (%s)" % (action, node_id, str(r.read()), r.status))
    except Exception:
        etype, value, tb = sys.exc_info()
        raise ShutdownError("Unable to contact CycleCloud while trying to %s node %s: %s" % (action, node_id, str(value)))

    # Go ahead and execute shutdown actions in case termination event is disabled
    shutdown_actions.handle_shutdown()


def request_shutdown(reason=None, config=None):
    """ Requests shutdown from CycleCloud 
    
    Returns True on success (200 OK), False otherwise
    """

    # TBD: we could try to execute shutdown actions here in case termination event is disabled,
    #      but if the terminate request fails, we could end up with a useless node
    # NOTE: actions would have to be executed before the request in order to ensure that the shutdown 
    #       actions complete before the node is terminated
    # shutdown_actions.handle_shutdown()

    configuration = jetpack.util.parse_config(config)

    cluster_name = configuration['identity']['cluster_name']
    instance_id = configuration['identity']['instance_id'] 

    url = urllib.parse.quote("/cloud/actions/autoscale/%s/stop" % cluster_name)
    params = {
        "instance": instance_id
    }
    if reason:
        params['reason'] = reason


    try:
        r = jetpack.util.query_cyclecloud(url, params=params, method="POST")
        if r.status != 200:
            raise ShutdownError("Unable to request node shutdown, received status %s from CycleCloud" % r.status)
    except:
        etype, value, tb = sys.exc_info()
        raise ShutdownError("Unable to connect to CycleCloud: %s" % str(value))
    

def shutdown():
    """ Issues a shutdown command to the OS """

    # Execute shutdown actions on manual shutdown
    shutdown_actions.handle_shutdown()

    if jetpack.util.is_windows():
        subprocess.check_call(['shutdown', '/s', '/f', '/t', '0'], shell=True)
    else:
        subprocess.check_call(['/sbin/poweroff'], shell=True)

