"""
Shutdown actions to be executed on Terminate, Preempt or OS Shutdown

NOTE:
- We currently DO NOT execute on Deallocate or Reboot since shutdown actions 
  may make the node unusable upon restart.
- In the future, we may want to allow registering actions with a list of conditions
  under which they execute...
"""
import logging
import jetpack.config
import jetpack.util
import jetpack.util.system
import subprocess
import os
import json
import shutil

logger = logging.getLogger()


class ShutdownActionHandler(jetpack.util.LockedFileManager):
    """ Stores the stack of registered shutdown actions in order of registration in a threadsafe way."""

    SHUTDOWN_ACTIONS_PATH = os.path.join(jetpack.util.jetpack_home(), "config", "shutdown_actions.json")

    def __init__(self):
        super().__init__()
        self._added_shutdown_actions = []

        # Create the directory if it doesn't exist  (generally only needed for tests)
        if not os.path.isdir(self.directory()):
            os.makedirs(self.directory(), exist_ok=True)

    def _remove_all(self):
        logger.info("Removing all registered shutdown actions...")
        self._added_shutdown_actions = []
        if os.path.exists(self.filename()):  
            os.remove(self.filename())

    def add_action(self, script_path):
        logger.info("Registering shutdown script: %s" % script_path)  
        self._added_shutdown_actions.append(script_path)        
        self.update()

    def execute(self):
        persisted_shutdown_actions = self._read()

        # Execute shutdown actions in reverse order of registration
        for script_path in reversed(persisted_shutdown_actions):
            try:
                logger.info("Executing shutdown script: %s" % script_path)
                cmd = ['/bin/bash', '-c', script_path]
                subprocess.check_call(cmd, shell=False)
                logger.info("Shutdown script completed successfully: %s" % ' '.join(cmd))
            except Exception as e:
                logger.exception("Error executing shutdown script: %s" % str(e))
                # Nothing to do here since the node is shutting down anyway...   Just move on to the next
                pass


    def filename(self) -> str:
        """ Get the node.json configuration based on the OS """
        return ShutdownActionHandler.SHUTDOWN_ACTIONS_PATH


    def directory(self) -> str:
        return os.path.dirname(self.filename())
    
    def check_for_update(self):
        # Always update
        return True;

    def _create_file_if_needed(self):
        # Should only be called from inside a locked write() method
        
        # Create the file if it doesn't exist 
        if not os.path.exists(self.filename()):  
            with open(self.filename(), 'w') as f:
                f.write(json.dumps([], indent=4))  
            dir_stats = os.stat(self.directory())
            jetpack.util.update_file_permissions(self.filename(), dir_stats.st_uid, dir_stats.st_gid, 0o660)


    def _read(self):      
        if not os.path.exists(self.filename()):
            return []
          
        with open(self.filename(), 'r') as f:
            return json.load(f)
        
    def write(self):
        # Atomic write called from locked update() method  
        # (read the persisted actions inside the lock to ensure consistency)

        self._create_file_if_needed()        
        persisted_shutdown_actions = self._read()

        # Add any shutdown actions that were added since the last update
        for shutdown_action in self._added_shutdown_actions:
            if shutdown_action not in persisted_shutdown_actions:
                persisted_shutdown_actions.append(shutdown_action)

        st = os.stat(self.filename())
        tmp_filename = self.filename() + ".tmp"
        with open(tmp_filename, 'w') as tmp:
            tmp.write(json.dumps(persisted_shutdown_actions, indent=4))

        jetpack.util.update_file_permissions(tmp_filename, st.st_uid, st.st_gid, 0o660)
        shutil.move(tmp_filename, self.filename())


def _reset_shutdown_actions():
    """Resets the shutdown actions to an empty list for testing purposes"""
    _shutdown_action_handler = ShutdownActionHandler()
    _shutdown_action_handler._remove_all()

def register_shutdown_script(script_path):    
    _shutdown_action_handler = ShutdownActionHandler()

    logger.info("Registering shutdown script: %s" % script_path)
    _shutdown_action_handler.add_action(script_path)

def get_registered_scripts():
    _shutdown_action_handler = ShutdownActionHandler()

    return _shutdown_action_handler._read()

def handle_shutdown():
    """Runs registered shutdown actions in reverse order of registration """
    _shutdown_action_handler = ShutdownActionHandler()

    logger.info("Running shutdown actions...")
    _shutdown_action_handler.execute()
