import os
import platform
import ctypes
import http.client
import datetime
from functools import wraps
import json
import time
import logging
import logging.config
import sys
import traceback
import subprocess
import base64
import urllib.request
import urllib.error
import urllib.parse
import re
import hashlib
import socket
import math
import ssl
from logging.handlers import RotatingFileHandler
from typing import Optional

JETPACK_USER_AGENT = "azure/jetpack"

import jetpack.util.system

logger = logging.getLogger(__name__)


class UserError(Exception):
    def __init__(self, message, description=None, detail=None, recommendation=None, help_link=None):
        """ Raise an exception which will be bubbled up to the CycleCloud UI

        message: Short message for use as a status line in datatable e.g. "Cluster-Init Error"
        description: One liner with more specific information for details pane e.g. "Error running cluster-init script install.sh"
        detail: Multiline description of the error e.g. stacktrace or stdout
        recommendation: Recommendation on how to resolve the error, if any
        help_link: URL to a specific help page about the error, if any
        """
        super(Exception, self).__init__(message)
        self.message = message
        self.description = description
        self.detail = detail
        self.recommendation = recommendation
        self.help_link = help_link
    
    def __str__(self):
        if not self. description and not self.detail:
            return self.message
        return f"Summary: {self.message}\nDescription: {self.description}\nDetail: {self.detail}".strip()
    

if os.name != "nt":
    # TODO: we may need a similar file-locking solution for Windows user support
    import fcntl


_unverified_context = ssl._create_unverified_context()
_unverified_context.options |= ssl.OP_NO_SSLv2 | ssl.OP_NO_SSLv3 | ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1 | ssl.OP_NO_COMPRESSION


def retry(tries: int, delay: int = 1, backoff: int = 1, description: Optional[str] = None):
    """ Decorator to retry a function 'tries' times with a delay of
    'delay' seconds with a scaling delay factor of 'backoff' between retries.
    """
    def retry_decorator(f):

        @wraps(f)
        def wrapper(*args, **kwargs):
            nonlocal description
            num_tries, current_delay = tries, delay
            while num_tries > 1:
                try:
                    return f(*args, **kwargs)
                except Exception as e:
                    if description is not None:
                        logger.warning("While {}, encountered {} retrying in {} seconds".format(description, e, current_delay))
                    else:
                        logger.warning("Encountered {} retrying in {} seconds".format(e, current_delay))
                    time.sleep(current_delay)
                    num_tries -= 1
                    current_delay *= backoff
            return f(*args, **kwargs)
        return wrapper
    return retry_decorator


def _timestamps_file():
    return os.path.join(jetpack_home(), 'config', 'timestamps.json')


def is_windows() -> bool:
    """ Is Jetpack running on a Win32 system """
    return platform.system().lower().startswith("win")


def is_linux() -> bool:
    """ Is Jetpack running on a Linux system """
    return not is_windows()


def version() -> str:
    ''' Returns the Jetpack version '''
    version_file = os.path.join(jetpack_home(), 'version.txt')
    version = None
    if os.path.exists(version_file):
        with open(version_file, 'r') as f:
            version = f.read().strip()
    else:
        raise Exception("Version file %s was not found" % version_file)

    return version


def set_cyclecloud_env() -> None:
    ''' Sets up the current python environment as the CycleCloud environment '''
    os.environ['CYCLECLOUD_HOME'] = jetpack_home()
    os.environ['CYCLECLOUD_BOOTSTRAP'] = os.path.join(jetpack_home(), 'system', 'bootstrap')

    def add_to_path(path: str) -> None:
        ''' Check to see if p is already in the system PATH and add it if not '''
        sep = ':'
        if is_windows():
            sep = ';'

        if path.lower() not in [p.lower() for p in os.environ['PATH'].split(sep)]:
            os.environ['PATH'] = sep.join([path] + os.environ['PATH'].split(sep))

    add_to_path(os.path.join(jetpack_home(), 'bin'))
    add_to_path(os.path.join(jetpack_home(), 'system', 'bin'))
    add_to_path(os.path.join(jetpack_home(), 'system', 'embedded', 'bin'))

    if is_windows():
        add_to_path(os.path.join(jetpack_home(), 'system', 'embedded', 'python'))
        add_to_path(os.path.join(jetpack_home(), 'system', 'embedded', 'python', 'scripts'))


def get_selinux_enforce() -> bool:
    # Selinux has different enforce flag paths for CentOS 6 and 7.
    if os.path.exists('/selinux/enforce'):
        with open('/selinux/enforce') as c6:
            return c6.read() == '1'
    elif os.path.exists('/sys/fs/selinux/enforce'):
        with open('/sys/fs/selinux/enforce') as c7:
            return c7.read() == '1'
    else:
        return False


def selinux_mode(mode=None):
    """ Sets the mode for selinux """

    if get_selinux_enforce():
        if mode not in ['permissive', 'disabled']:
            raise UserError("Unable to set SELinux mode", description="Unsupported SELinux mode specified: %s" % mode, detail="Available modes: permissive, disabled")

        jetpack.util.system.install_package('libselinux-utils')
        jetpack.util.system.execute('setenforce 0', shell=True)
        jetpack.util.system.execute("sed -i 's/^SELINUX=.*/SELINUX=%s/g' /etc/selinux/config" % mode, shell=True)

        msg = "SELinux set to permissive mode."
        if mode == 'disabled':
            msg += " A reboot is required to finishing disabling SELinux."
        logger.info(msg)
    else:
        logger.debug("SELinux is not currently in enforcing mode, nothing to do.")


def get_installer_dir():
    if is_windows():
        return 'C:\\cycle\\jetpack-installer'
    else:
        return '/opt/cycle/jetpack-installer'


def get_provider():
    return "azure"


@retry(tries=5, delay=2, description="querying instance metadata")
def _query_compute_metadata():
    """ Returns compute metadata as JSON """
    url = "http://169.254.169.254/metadata/instance/compute?api-version=2017-08-01"
    request = urllib.request.Request(url, None, {"Metadata": "true"})
    metadata = json.loads(urllib.request.urlopen(request).read())
    return metadata


def get_nodearray_name():
    try:
        metadata = _query_compute_metadata()
        # the part before the dash can be either a nodearray name or an instance name
        m = re.match(r"(.+)-(.+)", metadata["name"])
        if m:
            nodearray_name = m.group(1)
        else:
            nodearray_name = None
        return nodearray_name
    except Exception as e:
        raise ValueError("Unable to get nodearray name: %s" % str(e))


def get_instance_id():
    """Returns the instance ID for Azure. There is no need to process data from CycleCloud."""
    try:
        metadata = _query_compute_metadata()

        # the only indication given that this is a scaleset is that the name ends in _NNN
        m = re.match(r"(.+)_(\d+)", metadata["name"])
        if m:
            resource_id = "/subscriptions/%s/resourceGroups/%s/providers/Microsoft.Compute/virtualMachineScaleSets/%s/virtualMachines/%s" %  \
                          (metadata["subscriptionId"], metadata["resourceGroupName"], m.group(1), m.group(2))
        else:
            resource_id = "/subscriptions/%s/resourceGroups/%s/providers/Microsoft.Compute/virtualMachines/%s" % \
                          (metadata["subscriptionId"], metadata["resourceGroupName"], metadata["name"])
        logger.debug("Generating Azure ID using '%s'" % (resource_id))

        instance_id = hashlib.md5(resource_id.lower().encode('utf8')).hexdigest()
        return instance_id
    except Exception as e:
        raise ValueError("Unable to get instance_id: %s" % str(e))


def get_resource_name():
    try:
        metadata = _query_compute_metadata()
        return metadata['name']
    except Exception as e:
        raise ValueError("Unable to get resource name: %s" % str(e))


def pid_running(pid):
    """ Given a pid (int) check with the OS to see if it is currently running """
    if is_windows():
        kernel32 = ctypes.windll.kernel32
        SYNCHRONIZE = 0x100000

        process = kernel32.OpenProcess(SYNCHRONIZE, 0, pid)
        if process != 0:
            kernel32.CloseHandle(process)
            return True
        else:
            return False
    else:
        try:
            os.kill(pid, 0)
            return True
        except OSError:
            return False


def stale_file(pid_file):
    """ Check the last boot time vs the modified time of the pid file to see
    if it's a stale file """
    if not os.path.isfile(pid_file):
        return False

    last_modified = datetime.datetime.fromtimestamp(os.path.getmtime(pid_file))

    boottime = get_boottime()

    if boottime >= last_modified:
        return True
    else:
        return False


def get_boottime():
    if is_windows():
        kernel32 = ctypes.windll.kernel32
        kernel32.GetTickCount64.restype = ctypes.c_uint64
        uptime = kernel32.GetTickCount64() / 1000.0
    else:
        with open('/proc/uptime', 'r') as f:
            uptime = float(f.readline().split()[0])

    boottime = datetime.datetime.fromtimestamp(time.time() - uptime)
    return boottime


def already_running(pid_file=None):
    """ Given a path to a pid file, check to see if the pid in the file is running:
      * if yes - return True as the process is already running
      * if no - return False as the process is no longer running
      * if no pid file exsits - write one for the current process
     """
    pid = None

    # Read the pid from the file if it exists and isn't stale
    if os.path.isfile(pid_file) and not stale_file(pid_file):
        with open(pid_file) as f:
            try:
                pid = int(f.read())
            except ValueError:
                # If we got a value error the pid file contains something other than a
                # process ID, thus we can assume that the process isn't running
                return False

    # If no pid file was read, or it was read and the process is no longer running
    # return false and write the current pid to the file
    if pid is None or (pid and not pid_running(pid)):
        with open(pid_file, 'w') as f:
            f.write(str(os.getpid()))
        return False

    # If the pid exists and it is running, return True - nothing to see here
    else:
        return True


def _config_to_dict(config_parser):
    # Convert parser to dictionary
    cfg = {}
    for section_name in config_parser.sections():
        section = {}
        for option in config_parser.options(section_name):
            section[option] = config_parser.get(section_name, option)
        cfg[section_name] = section

    return cfg


def _get_config_file(config=None):
    if config is None:
        config = '/opt/cycle/jetpack/config/jetpack.ini'
        if is_windows():
            config = 'C:\\cycle\\jetpack\\config\\jetpack.ini'
    config = os.path.expanduser(config)
    return config


def _transform_dict_to_ini(config_dict):
    """ Converts a config dictionary into an .ini file """
    def _format_section(name, section):
        section = '\n'.join(["%s=%s" % (k, v) for k, v in section.items()])
        return "[%s]\n%s" % (name, section)

    config_ini = '\n\n'.join([_format_section(k, v) for k, v in config_dict.items()])

    return config_ini


# Handler to set ownership for the rotated log files
# See python docs: http://python.readthedocs.org/en/latest/howto/logging-cookbook.html
def owned_file_handler(filename, owner=None):
    if owner and not is_windows():
        if not os.path.exists(filename):
            prv = os.umask(0o002)
            open(filename, 'a').close()
            os.umask(prv)
        return OwnedRotatingFileHandler(filename=filename, backupCount=5, maxBytes=10485740, owner=owner)
    else:
        return RotatingFileHandler(filename, backupCount=5, maxBytes=10485740)


def setup_logging(loglevel="INFO"):
    cyclecloud_home = jetpack_home()
    logfile = os.path.join(cyclecloud_home, 'logs', 'jetpack.log')

    LOGGING = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'default': {
                'format': '%(asctime)s %(levelname)-8s %(message)s'
            },
        },
        'handlers': {
            'file': {
                # The values below are popped from this dictionary and
                # used to create the handler, set the handler's level and
                # its formatter.
                '()': DirOwnedRotatingFileHandler,
                'level': loglevel.upper(),
                'mode': 'a',
                'formatter': 'default',
                # The values below are passed to the handler creator callable
                # as keyword arguments.
                'filename': logfile,
            },
        },
        'root': {
            'handlers': ['file'],
            'level': loglevel.upper(),
        },
    }

    logging.config.dictConfig(LOGGING)


def parse_config(config=None):
    config = _get_config_file(config)
    if not os.path.isfile(config):
        return {}

    # Now read the config file
    import configparser
    parser = configparser.RawConfigParser()
    with open(config, 'r') as f:
        parser.read_file(f, config)

    return _config_to_dict(parser)


def write_config(config_dict, config=None):
    config = _get_config_file(config)
    config_ini = _transform_dict_to_ini(config_dict)

    config_dir = os.path.dirname(config)
    if not os.path.exists(config_dir):
        os.makedirs(config_dir)

    with open(config, 'w') as f:
        f.write(config_ini)
    if not is_windows():
        os.chmod(config, 0o660)


def iso_8601_timestamp():
    return time.strftime('%Y-%m-%dT%H:%M:%S+00:00', time.gmtime())


def get_connection_config():
    """ Reads and returns the connection.json file """
    with open(os.path.join(jetpack_home(), 'config', 'connection.json')) as f:
        return json.load(f)


def get_cyclecloud_connection():
    """ Gets a connection to CycleCloud

    Returns a tuple containing:
     - Http(s)Connection object
     - headers for the request (http authorization set)
     """
    config = get_connection_config()

    url = config.get('url')
    username = config.get('username')
    password = config.get('password')

    conn = None
    u = urllib.parse.urlparse(url)
    if u[0].lower() == "https":
        if 'cert' in config:
            # using cadata in this helper function makes sure the server certificate matches this cert only, not any system trusted certificate
            context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH, cadata=config['cert'])
            context.options |= ssl.OP_NO_SSLv2 | ssl.OP_NO_SSLv3 | ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1 | ssl.OP_NO_COMPRESSION
            context.verify_mode = ssl.CERT_REQUIRED
            # note: we don't check the hostname because that would require us writing to /etc/hosts or
            # providing some other way for DNS resolution to work.
            context.check_hostname = False

        else:
            # Note: You can not pass in a cert by setting `cyclecloud.skip_ssl_validation=True` in node configuration
            context = _unverified_context

        conn = http.client.HTTPSConnection(u[1], context=context)
    else:
        conn = http.client.HTTPConnection(u[1])

    auth = 'Basic ' + base64.b64encode(b':'.join([username.encode('utf8'), password.encode('utf8')]).strip()).decode('utf8')
    headers = {
        'Authorization': auth,
        'User-Agent': JETPACK_USER_AGENT
    }

    return conn, headers


# how often (in seconds) to attempt a connection (this is also the timeout)
SOCKET_TIMEOUT = 10

# how long in seconds to wait, in total
TOTAL_ATTEMPT_TIME = 120

# how long to wait if going through a proxy (longer to allow the tunnel to be set up)
TOTAL_ATTEMPT_TIME_THROUGH_PROXY = 300


def test_cyclecloud_connection():
    '''Attempts to communicate with CycleCloud. Raises an exception if it cannot, returns nothing if it can.'''
    connection = get_connection_config()

    instance_id = get_instance_id()

    path = "/cloud/api/node_status/%s?cluster=%s" % (urllib.parse.quote(instance_id), urllib.parse.quote(connection["cluster"]))
    parsed_url = urllib.parse.urlparse(connection["url"])
    parts = parsed_url.netloc.split(":")
    if len(parts) == 2:
        host = parts[0]
        port = parts[1]
    else:
        host = parts[0]
        port = "443" if parsed_url.scheme == "https" else "80"

    port = int(port)

    formatted_url = connection["url"]

    error = None
    detail = None

    recommendation = None

    total_time = TOTAL_ATTEMPT_TIME
    if connection.get("proxy_method") in ["local", "remote"]:
        total_time = TOTAL_ATTEMPT_TIME_THROUGH_PROXY

    connection_attempts = int(math.ceil(float(total_time) / SOCKET_TIMEOUT))

    dest_label = "CycleCloud at %s" % formatted_url
    if connection.get("proxy_method") in ["local", "remote"]:
        dest_label = "CycleCloud through return-proxy tunnel at %s" % formatted_url
        # TODO: this could be more detailed
        recommendation = "Make sure CycleCloud has established a return-proxy tunnel."

    for x in range(connection_attempts):
        start = time.time()
        try:
            r = _connect_to_cyclecloud(path)
            if r.status != 200:
                error = "Unexpected error connecting to CycleCloud at %s: status code %s" % (formatted_url, r.status)
            else:
                # done
                error = None
                break

        except socket.timeout:
            error = "Timed out connecting to %s" % dest_label
        except ssl.SSLError as e:
            if hasattr(e, "reason") and e.reason == "CERTIFICATE_VERIFY_FAILED":
                error = "Certificate validation failed for CycleCloud at %s" % (formatted_url)
            elif hasattr(e, "message") and str(e).find("handshake operation timed out") != -1:
                # you get this when connecting through an ssh tunnel which is up, but which cannot connect to CycleServer.
                # in that case the exception does NOT have a reason or non-None errno and only gives a message like:
                #     _ssl.c:710: The handshake operation timed out
                # so we special-case it here. Note the wording is different than the above message so we can tell which it is.
                error = "Connection timed out to %s" % dest_label
            else:
                error = "Could not connect to %s: %s" % (dest_label, e)
        except socket.error as e:
            if e.errno == 111:
                # this is common ("connection refused") so we special-case the message
                error = "Connection refused to %s" % dest_label
            else:
                # this is a generic error message to handle all the other cases
                error = "Could not connect to %s: %s" % (dest_label, e)

        remainder = max(0, SOCKET_TIMEOUT - (time.time() - start))
        time.sleep(remainder)

    if error:
        raise UserError(error, recommendation=recommendation, detail=detail)


def query_cyclecloud(path: str, body=None, headers=None, params=None, method="GET") -> http.client.HTTPResponse:
    """ Queries cyclecloud with a HTTP(S) request """
    conn, default_headers = get_cyclecloud_connection()
    conn.timeout = SOCKET_TIMEOUT

    config = get_connection_config()

    # In some cases CycleCloud isn't located at '/' but instead soemthing like '/cyclecloud' 
    # so we have to ensure that we prepend this to the request path  dfor the caller
    url = config.get('url')
    u = urllib.parse.urlparse(url)
    context_path = u[2]  # '' or '/something'

    # Update the default headers from cyclecloud connectinon with user specified headers, overwriting the defaults
    if headers:
        default_headers.update(headers)

    # Add context path and encode params into
    if context_path and context_path != "/":
        path = "{}{}".format(context_path, path)
    if params:
        path = "{}?{}".format(path, urllib.parse.urlencode(params))

    conn.request(method, path, body=body, headers=default_headers)

    return conn.getresponse()


def get_current_config(config=None):
    if config is None:
        config = {}
    configuration = parse_config(None)

    connection_config = get_connection_config()

    instance_id = get_instance_id()

    params = {}
    params['cluster'] = connection_config.get('cluster', '')
    if 'identity' in configuration:
        params['template'] = configuration['identity']['template']
        params['cluster_session_id'] = configuration['identity']['cluster_session_id']
    else:
        # this should only happen before the instance is mapped to the node
        params['template'] = get_nodearray_name()

    url = "/cloud/api/node_status/%s" % urllib.parse.quote(instance_id)

    logger.debug("Getting node status from %s" % url)

    r = query_cyclecloud(url, params=params)
    if r.status != 200:
        logger.error("An error occurred getting node configuration %s: %s (%s)" % (url, r.reason, r.status))
        raise Exception(r.reason)

    body = json.loads(r.read())

    # write the node name/id etc to the config section
    cyclecloud_config = config.get("cyclecloud", {})
    config["cyclecloud"] = cyclecloud_config
    node_config = cyclecloud_config.get("node", {})
    cyclecloud_config["node"] = node_config
    node_config["ready"] = body["ready"]
    if "nodearray_parameters" in body:
        node_config["nodearray_parameters"] = body["nodearray_parameters"]

    if "name" in body:
        node_config["name"] = body["name"]
        node_config["id"] = body["id"]
        node_config["session_id"] = body["session_id"]

    logger.debug("Node configurations received: " + str(config))
    
    # TODO: extract configuration from body if present
    return config


def _connect_to_cyclecloud(path: str) -> http.client.HTTPResponse:
    return query_cyclecloud(path)


def deep_merge(src, dst):
    for k in src.keys():
        if k in dst and isinstance(src[k], dict) and isinstance(dst[k], dict):
            deep_merge(src[k], dst[k])
        else:
            dst[k] = src[k]


def jetpack_home() -> str:
    """ Returns the home/root directory where Jetpack is installed """
    if is_windows():
        return "C:\\cycle\\jetpack"
    else:
        return "/opt/cycle/jetpack"


def has_converged():
    """ Returns True/False depending on if the initial converge has completed """
    timestamps = {}

    # If the file already exists, read it
    if os.path.isfile(_timestamps_file()):
        with open(_timestamps_file(), 'r') as f:
            timestamps = json.load(f)

    # The node has converged if there is a 'converge_initial_finish_timestamp' key
    return 'converge_initial_finish_timestamp' in timestamps


def update_file_permissions(path, uid=-1, gid=-1, mode=None):
    """ Update a file/directory with the appropriate UID/GID and permissions """
    if is_windows():
        return
    file_stat = os.stat(path)
    if uid != -1 or gid != -1:
        if file_stat.st_uid != uid or file_stat.st_gid != gid:
            os.chown(path, uid, gid)
    if mode is not None and (mode & file_stat.st_mode) != mode:
        os.chmod(path, mode)


def write_timestamps(timestamps):
    """ Writes the timestamps dictionary to a json file and return the contents of the file """

    existing = {}

    # Convert the passed in timestamps to ISO-8601 format
    for k, v in timestamps.items():
        timestamps[k] = v.isoformat()

    # If the file already exists, read it
    if os.path.isfile(_timestamps_file()):
        with open(_timestamps_file(), 'r') as f:
            existing = json.load(f)

    # Merge the existing timestamps on top of the new timestamps
    # i.e. if they already existed, keep the original values
    deep_merge(existing, timestamps)

    # Write it back to disk
    with open(_timestamps_file(), 'w') as f:
        f.write(json.dumps(timestamps, sort_keys=True))

    return timestamps


def separate(path):
    """ Separates a path into all it's component parts

    Example: "/path/to/file" => ['path', 'to', 'file']

    Note: Opposite of os.path.join() you round trip like: `os.path.join(*separate('/foo/bar/baz'))=='/foo/bar/baz'`
    """
    head, tail = os.path.split(path)
    if head == path:
        return [head]
    return separate(head) + [tail]


def commonpath(paths):
    """ Give a list of paths, find the longest comment path between them

    Example: ['/shared/home', '/shared'] => '/shared
    ['/home', '/shared/home']=> '/'

    Note: Similar to the Python3 method that is missing in Python2
    """
    common = None
    for path in paths:
        parts = separate(path)
        if common is None:
            common = parts
        else:
            new = []
            for x, y in zip(common, parts):
                if x == y:
                    new.append(x)
            common = new
    if common:
        return os.path.join(*common)
    else:
        return ''


def run_once(f):
    """ Run a function once to completion, exceptions and other errors will allow it to be rerun

    @run_once
    def add(x, y):
        return x + y

    add(2, 4) => 6
    add(3, 3) => None  (didn't run)
    """
    def wrapper(*args, **kwargs):
        if not wrapper.complete:
            r = f(*args, **kwargs)
            wrapper.complete = True
            return r
    wrapper.complete = False
    return wrapper


def rebuild_rpmdb():
    """ Check to see if the RPMDB is corrupt and rebuild it if necessary """
    rebuilt = False
    try:
        if os.path.exists('/etc/redhat-release'):
            try:
                subprocess.check_output(['rpm', '-qa'], stderr=subprocess.STDOUT, text=True)
            except subprocess.CalledProcessError as rpmerr:
                if rpmerr.returncode == 1 and rpmerr.output is not None and "run database recovery" in rpmerr.output.lower():
                    logger.warning("Detected RPM database corruption, attempting to repair")
                    subprocess.check_call(['rpm', '--rebuilddb'])
                    logger.info("Rebuilt corrupted RPM database")
                    rebuilt = True
    except Exception:
        logger.exception("Error handling RPM database corruption")

    return rebuilt


class LockedFileManagerException(Exception):
    pass


class LockedFileManager(object):
    """ Locks a directory (via .jetpack.lock file) so that a sublcass can modify files in that class without
    worry of being clobbered by another instance implementing LockedFileManager. This is useful for preventing multiple
    nodes that write to the same shared filesystem from clobbering the same file """

    def directory(self):
        """ Return the directory that we'll be writing to """
        raise NotImplementedError("Sublcass must implement this method")

    def check_for_update(self):
        """ This method should check if the resource needs an update (True/False) """
        raise NotImplementedError("Subclass must implement this method")

    def write(self):
        """ This method should actually write the resource to disk """
        raise NotImplementedError("Subclass must implement this method")

    def update(self, max_attempts=30):
        """ Check and then update the resource on disk as needed using locking to prevent multiple jetpack processes from writing to the file
        at the same time. This method does a lock on the directory via .jetpack.lock file, not the resource in question

        Returns: True/False if the resource was modified
        """
        attempts = 0
        lockfile = os.path.join(self.directory(), ".jetpack.lock")
        modified = False
        while attempts < max_attempts:
            # Check to see if we even need to perform an update, if not we can just move on
            if not self.check_for_update():
                break
            with open(lockfile, 'w') as f:
                try:
                    fcntl.lockf(f, fcntl.LOCK_EX | fcntl.LOCK_NB)
                except IOError as e:
                    # Errno 11 is "Resource temporarily unavailable" i.e. the file is locked already
                    if e.errno == 11:
                        attempts += 1
                        time.sleep(1)
                        continue
                    else:
                        raise

                # change ownership of the file to match the directory
                dir_stats = os.stat(self.directory())
                lock_stats = os.fstat(f.fileno())
                if lock_stats.st_uid != dir_stats.st_uid or lock_stats.st_gid != dir_stats.st_gid:
                    os.fchown(f.fileno(), dir_stats.st_uid, dir_stats.st_gid)

                # Re-check to see if we still need an update, if so, write it inside since we have the lock
                if self.check_for_update():
                    self.write()
                    modified = True
                break
        # Raise an exception if we hit our max attempts
        if attempts == max_attempts:
            raise LockedFileManagerException("Maximum attempts reached trying to write to locked file")

        return modified


def _configuration_status_file():
    return os.path.join(jetpack_home(), 'system', 'run', 'configuration-status.json')


def clear_configuration_status():
    """ Removes the converge status file at the beginning of the chef-run """
    if os.path.isfile(_configuration_status_file()):
        os.remove(_configuration_status_file())


def read_configuration_status():
    """ Reads the configuration-status.json file from disk, this file is
    used to track errors that happen during converge (i.e. talk between python and ruby processes) """
    status = {}
    if os.path.isfile(_configuration_status_file()):
        try:
            with open(_configuration_status_file()) as f:
                status = json.loads(f.read())
        except Exception as e:
            status = {
                'description': 'Unable to read configuration-status.json',
                'detail': str(e)
            }
    return status


def write_configuration_exception(e, force=False):
    """ Writes an exception to the configuration-status.json file on disk, only for errors during Chef converge """
    status = {}
    if isinstance(e, UserError):
        status['message'] = str(e)
        status['description'] = e.description
        status['detail'] = e.detail
        status['recommendation'] = e.recommendation
        status['help_link'] = e.help_link
    elif isinstance(e, Exception):
        # In the case of a general exception there is no message, we'll just let the handler decide how to classify
        # it. The description of the state is the exception message and the detail will be the traceback (if available).
        status['description'] = str(e)

        # Attempt to extract a traceback if one is available and put it in the detail
        ex_type, ex, tb = sys.exc_info()
        if tb:
            status['detail'] = "".join(traceback.format_tb(tb))

    # Only update the file on disk if it doesn't already exist (or force==True)
    # If it does exist that means some other process has caught a more detailed error earlier
    # and already wrote it to disk
    if not os.path.isfile(_configuration_status_file()) or force is True:
        with open(_configuration_status_file(), 'w') as f:
            f.write(json.dumps(status, indent=4, sort_keys=True))


# this is needed for the logs to require less privilege to access than jetpack needs to run
class OwnedRotatingFileHandler(RotatingFileHandler):
    def __init__(self, filename, backupCount=5, maxBytes=10485740, owner=['root', 'cyclecloud'], mode=0o660):
        # have to call this before the super constructor due to "object has no attribute 'owner'" otherwise
        self.owner = owner
        self.filemode = mode
        super().__init__(filename, backupCount, maxBytes)
            
    def _open(self):
        # change ownership of the file to match the directory
        logfile = RotatingFileHandler._open(self)
        if not is_windows():
            import grp
            import pwd
            uid = pwd.getpwnam(self.owner[0]).pw_uid
            gid = grp.getgrnam(self.owner[1]).gr_gid
            update_file_permissions(self.baseFilename, uid, gid, self.filemode)
        return logfile


class DirOwnedRotatingFileHandler(RotatingFileHandler):
    def _open(self):
        # change ownership of the file to match the directory
        logfile = RotatingFileHandler._open(self)
        if not is_windows():
            dir_stats = os.stat(os.path.dirname(self.baseFilename))
            try:
                update_file_permissions(self.baseFilename, dir_stats.st_uid, dir_stats.st_gid, mode=0o660)
            except Exception as ex:
                print(f"Could not change file permissions for {self.baseFilename}: {ex}", stream=sys.stderr)
        return logfile
