import logging
import os
import shutil
import subprocess

import jetpack.util

logger = logging.getLogger(__name__)

class SystemUtilityError(Exception):
    pass

def platform_family():
    """ Determine the platform family currently being run """
    if jetpack.util.is_windows():
        return 'windows'
    elif os.path.exists('/etc/redhat-release'):
        return 'rhel'
    elif os.path.exists('/etc/debian_version'):
        return 'debian'
    else:
        raise SystemUtilityError("Unable to determine platform family")

def install_package(name):
    """ Install a package NAME if it isn't already installed """
    if not package_already_installed(name):
        try:
            if platform_family() == "rhel":
                subprocess.check_output(['yum', 'install', '-y', name], stderr=subprocess.STDOUT)
            elif platform_family() == "debian":
                subprocess.check_output(['apt-get', 'install', '-y', name], stderr=subprocess.STDOUT)
            else:
                raise SystemUtilityError("Unable to install package: not a rhel or debian based platform")
        except subprocess.CalledProcessError as e:
            logger.exception("Unable to install package %s" % name)
            raise SystemUtilityError("Unable to install package %s: %s (%s)" % (name, e.output, e.returncode))
        
        logger.info("Installed package '%s'" % name)
    else:
        logger.debug("Package '%s' already installed, nothing to do." % name)

def package_already_installed(name):
    """ Check to see if a package is already installed """
    try:
        if platform_family() == "rhel":
            subprocess.check_output(['rpm', '-q', name])
        elif platform_family() == 'debian':
            subprocess.check_output(['dpkg', '-l', name])
        else:
            raise SystemUtilityError("Unable to check package status: not a rhel or debian based platform")
    except subprocess.CalledProcessError as e:
        return False
    return True
    
def is_systemd():
    """ Check to see if the Linux distro supports systemd by reading /sbin/init """
    if jetpack.util.is_linux() and os.path.islink("/sbin/init"):
        if "systemd" in os.readlink("/sbin/init"):
            return True
    return False

def windows_env(key, value):
    ''' Set a Windows environment key/value pair '''
    try:
        with open(os.devnull, 'w') as devnull:
            subprocess.check_call("setx /M {0} \"{1}\"".format(key, value), shell=True, stdout=devnull)
        logger.info("Set ENV variable %s='%s'" % (key, value))
    except subprocess.CalledProcessError as e:
        raise SystemUtilityError("Unable to set Windows env %s='%s': %s (%s)" % (key, value, e.output, e.returncode))

def execute(cmd, shell=False, msg=None):
    ''' Execute a CMD'''
    try:
        with open(os.devnull, 'w') as devnull:
            subprocess.check_call(cmd, shell=shell, stdout=devnull)
        if msg:
            logger.info(msg)
        else:
            logger.info("Ran command: %s" % cmd)
    except subprocess.CalledProcessError as e:
        raise SystemUtilityError("Unable to execute command '%s': %s (%s)" % (cmd, e.output, e.returncode))

def service(name, enable=False, start=False):
    """ Enable/Start services in Linux """
    systemd = is_systemd()
    try:
        if enable:
            if jetpack.util.is_windows():
                with open(os.devnull, 'w') as devnull:
                    subprocess.check_call(['sc', 'config', name, 'start=auto'], stdout=devnull)
            elif systemd:
                with open(os.devnull, 'w') as devnull:
                    subprocess.check_call(['systemctl', 'enable', name], stdout=devnull)
            else:
                # When using sysvinit we have two ways of enabling the service (update-rc.d on Debian and chkconfig on RHEL)
                if shutil.which("update-rc.d"):
                    with open(os.devnull, 'w') as devnull:
                        subprocess.check_call(['update-rc.d', name, 'defaults'], stdout=devnull)
                else:
                    with open(os.devnull, 'w') as devnull:
                        subprocess.check_call(['chkconfig', '--add', name], stdout=devnull)
            logger.info("Enabled %s service" % name)
    except subprocess.CalledProcessError as e:
        raise SystemUtilityError("Failed to enable the %s service: %s (%s)" % (name, e.output, e.returncode))

    try:
        if start:
            if jetpack.util.is_windows():
                with open(os.devnull, 'w') as devnull:
                    try:
                        subprocess.check_call(['sc', 'start', name], stdout=devnull)
                    except subprocess.CalledProcessError as e:
                        # If the service is already running a 1056 will be returned which is an OK status, otherwise reraise
                        if e.returncode != 1056:
                            raise
            elif systemd:
                with open(os.devnull, 'w') as devnull:
                    subprocess.check_call(['systemctl', 'start', name], stdout=devnull)
            else:
                with open(os.devnull, 'w') as devnull:
                    subprocess.check_call(['/etc/init.d/%s' % name, 'start'], stdout=devnull)
            logger.info("Started %s service" % name)
    except subprocess.CalledProcessError as e:
        raise SystemUtilityError("Failed to start the %s service: %s (%s)" % (name, e.output, e.returncode))

def create_group(name, gid=None):
    """ Create a group NAME with a given gid """

    try:
        # Check to see if the group exists, if it does just return
        import grp
        grp.getgrnam(name)
        logger.warning("Unable to add group '%s', it already exists." % name)
        return
    except KeyError:
        pass # We expect a KeyError when we are creating a group

    cmd = ['groupadd', name]
    if gid:
        cmd += ['--gid', '%s' % gid]
    
    try:
        subprocess.check_call(cmd)
        msg = "Created group '%s'" % name
        if gid:
            msg += " (gid: %s)" % gid
        logger.info(msg)
    except subprocess.CalledProcessError as e:
        raise SystemUtilityError("Unable to add group '%s': %s (%s)" % (name, e.output, e.returncode))

def create_user(name, group=None, uid=None, home_dir=None, shell=None):
    """ Create a new user NAME with given parameters """
    try:
        import pwd
        pwd.getpwnam(name)
        logger.warning("Unable to add user '%s', it already exists." % name)
        return
    except KeyError:
        pass  # Expected when the user doesn't exist and should be created

    cmd = ['useradd', name, '--create-home']

    if group:
        cmd += ['--gid', group]

    if uid:
        cmd += ['--uid', uid]

    if home_dir:
        cmd += ['--home', home_dir]

    if shell:
        cmd += ['--shell', shell]
    
    try:
        subprocess.check_call(cmd)
        logger.info("Created user '%s'" % name)
    except subprocess.CalledProcessError as e:
        raise SystemUtilityError("Unable to add user '%s': %s (%s)" % (name, e.output, e.returncode))

def copy_file(src, path, owner=None, group=None, permissions=None, mkdirs=True):
    """ Copy a file named NAME to PATH with OWNER:GROUP and PERMISSIONS """

    d = os.path.dirname(path)
    if mkdirs and not os.path.isdir(d):
        os.makedirs(d)

    shutil.copy(src, path)

    logger.info("Copied file: %s" % path)

    # Set the OWNER:GROUP and PERMISSIONS
    modify_file(path, owner, group, permissions)


def modify_file(path, owner=None, group=None, permissions=None):
    """ Modify the owner/group and permissions of a file if linux """
    if not jetpack.util.is_linux():
        logger.warning("Could not modify owner/group/permissions for `%s` because the OS is not Linux" % path)
        return

    import grp
    import pwd

    owner_id = owner
    group_id = group

    if owner_id is None:
        owner_id = -1
    elif isinstance(owner_id, str):
        try:
            owner_id = pwd.getpwnam(owner)[2]
        except KeyError:
            raise SystemUtilityError("Unable to determine uid for user '%s'" % owner_id)

    if group_id is None:
        group_id = -1
    elif isinstance(group_id, str):
        try:
            group_id = grp.getgrnam(group_id)[2]
        except KeyError:
            raise SystemUtilityError("Unable to determine gid for group '%s'" % group_id)
    
    if owner_id != -1 and group_id != -1:
        os.chown(path, owner_id, group_id)
        logger.info("Changed file ownership for: %s (%s:%s)" % (path, owner, group))
    if permissions:
        os.chmod(path, permissions)
        logger.info("Changed file permissions for %s (0x%o)" % (path, permissions))