import os
import json
from jetpack import cluster_init, converge
from jetpack import config as jetpack_config
import pytest  # type: ignore
from unittest import mock


def test_run_one_spec_good(datadir, mock_results_filepath):
    with mock.patch('jetpack.cluster_init._get_cluster_init_home', return_value=os.path.join(datadir, 'onespec')):
        with mock.patch('jetpack.cluster_init.get_specs') as mock_get_specs:
            mock_get_specs.return_value = [
                cluster_init.ClusterInitSpec({'project': 'htcondor', 'version': '1.0.0', 'project_spec': 'default'})
            ]
    
            report = cluster_init.run_tests(send_report=False)
            message = report.to_message()
            assert message['summary']['num_tests'] == 2
            assert message['summary']['failed'] == 1
            assert message['summary']['passed'] == 1
            assert len(message['tests']) == 2
            first_test = message['tests'][0]
            print(first_test)
            
            assert first_test['call']['outcome'] == 'failed'
            assert first_test['project'] == 'htcondor'
            assert first_test['spec'] == 'default'


def test_run_one_spec_notests(datadir, mock_results_filepath):
    with mock.patch('jetpack.cluster_init._get_cluster_init_home', return_value=os.path.join(datadir, 'notests')):
        with mock.patch('jetpack.cluster_init.get_specs') as mock_get_specs:
            mock_get_specs.return_value = [
                cluster_init.ClusterInitSpec({'project': 'htcondor', 'version': '1.0.0', 'project_spec': 'default'})
            ]
            report = cluster_init.run_tests(send_report=False)
            message = report.to_message()
            assert message['summary']['num_tests'] == 0
            assert message['summary']['failed'] == 0
            assert len(message['tests']) == 0
            assert message['status'] == 'notests'


def test_run_one_spec_skipped(datadir, mock_results_filepath):
    with mock.patch('jetpack.cluster_init._get_cluster_init_home', return_value=os.path.join(datadir, 'skipped')):
        with mock.patch('jetpack.cluster_init.get_specs') as mock_get_specs:
            mock_get_specs.return_value = [
                cluster_init.ClusterInitSpec({'project': 'htcondor', 'version': '1.0.0', 'project_spec': 'default'})
            ]
            report = cluster_init.run_tests(send_report=False)
            message = report.to_message()
            assert message['summary']['num_tests'] == 5
            assert message['summary']['failed'] == 0
            assert message['summary']['skipped'] == 5
            assert len(message['tests']) == 5
            assert message['status'] == 'skipped'


def test_run_multiple_specs_bad(datadir, mock_results_filepath):
    with mock.patch('jetpack.cluster_init._get_cluster_init_home', return_value=os.path.join(datadir, 'multiplespecs')):
        with mock.patch('jetpack.cluster_init.get_specs') as mock_get_specs:
            mock_get_specs.return_value = [
                cluster_init.ClusterInitSpec({'project': 'htcondor', 'version': '1.0.0', 'project_spec': 'default'}),
                cluster_init.ClusterInitSpec({'project': 'htcondor', 'version': '1.0.0', 'project_spec': 'superduper'})
            ]
            report = cluster_init.run_tests(send_report=False)
            message = report.to_message()
            failed_tests = [t for t in message['tests'] if t['outcome'] in ['failed']]

            # the failure message should only have the actual failure message and not the full traceback
            assert failed_tests[0]['call']['longrepr'].startswith('AssertionError: Big ugly failure')

            assert message['summary']['num_tests'] == 4
            assert message['summary']['failed'] == 3
            assert message['summary']['passed'] == 1
            assert message['status'] == 'failed'

def test_run_multiple_specs_notests(datadir, mock_results_filepath):
    with mock.patch('jetpack.cluster_init._get_cluster_init_home', return_value=os.path.join(datadir, 'multiplespecs-notests')):
        with mock.patch('jetpack.cluster_init.get_specs') as mock_get_specs:
            mock_get_specs.return_value = [
                cluster_init.ClusterInitSpec({'project': 'htcondor', 'version': '1.0.0', 'project_spec': 'default'}),
                cluster_init.ClusterInitSpec({'project': 'htcondor', 'version': '1.0.0', 'project_spec': 'superduper'})
            ]
            report = cluster_init.run_tests(send_report=False)
            message = report.to_message()
            assert message['summary']['num_tests'] == 0
            assert message['status'] == 'notests'

def test_with_multiple_specs_run_single(datadir, mock_results_filepath):
    with mock.patch('jetpack.cluster_init._get_cluster_init_home', return_value=os.path.join(datadir, 'multiplespecs')):
        with mock.patch('jetpack.cluster_init.get_specs') as mock_get_specs:
            mock_get_specs.return_value = [
                cluster_init.ClusterInitSpec({'project': 'htcondor', 'version': '1.0.0', 'project_spec': 'default'})
            ]
            report = cluster_init.run_tests(send_report=False)
            message = report.to_message()
            assert message['summary']['num_tests'] == 2
            assert message['summary']['failed'] == 1
            assert message['summary']['passed'] == 1
            assert message['status'] == 'failed'
    

def test_pytest_wrapper(datadir):
    specdir = os.path.join(datadir, 'onespec/htcondor/default')
    project = 'htcondor'
    spec = 'default'
    status, report = cluster_init.pytest_wrapper.run(specdir, project, spec)
    completed_tests = report['report']['tests']
    first_test = completed_tests[0]
    assert project == first_test['project']
    assert spec == first_test['spec']
    assert 'failed' == first_test['call']['outcome']

    summary = report['report']['summary']
    assert 1 == summary['failed']
    assert 2 == summary['num_tests']


def test_pytest_wrapper_collection_error(datadir):
    specdir = os.path.join(datadir, 'import_error/htcondor/default')
    project = 'htcondor'
    spec = 'default'
    status, report = cluster_init.pytest_wrapper.run(specdir, project, spec)
    completed_tests = report['report']['tests']
    error_test = completed_tests[0]
    assert project == error_test['project']
    assert spec == error_test['spec']
    assert 'error' == error_test['setup']['outcome']
    assert error_test['setup']['longrepr'].startswith('ImportError while importing test module')

    summary = report['report']['summary']
    assert 1 == summary['error']
    assert 1 == summary['num_tests']


def test_pytest_wrapper_syntax_error(datadir):
    specdir = os.path.join(datadir, 'syntax_error/htcondor/default')
    project = 'htcondor'
    spec = 'default'
    status, report = cluster_init.pytest_wrapper.run(specdir, project, spec)
    completed_tests = report['report']['tests']
    error_test = completed_tests[0]
    assert project == error_test['project']
    assert spec == error_test['spec']
    assert 'error' == error_test['setup']['outcome']
    assert 'SyntaxError: invalid syntax' in error_test['setup']['longrepr']

    summary = report['report']['summary']
    assert 1 == summary['error']
    assert 1 == summary['num_tests']


def test_pytest_wrapper_no_spec(datadir):
    nosuchspecdir = os.path.join(datadir, 'spec-test-does-not-exist')
    project = 'htcondor'
    spec = 'notdefault'
    with pytest.raises(cluster_init.pytest_wrapper.SpecNotFound):
        cluster_init.pytest_wrapper.run(nosuchspecdir, project, spec)

    
@pytest.fixture
def datadir():
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data')


@pytest.fixture
def mock_results_filepath(tmpdir, monkeypatch):
    monkeypatch.setattr(cluster_init, '_get_results_filepath',
                        lambda: os.path.join(str(tmpdir), 'test-results.json'))

