import pytest  # type: ignore
from unittest import mock
from typing import Optional

from jetpack.lockers import Locker
from jetpack.util import azcopy


@pytest.mark.parametrize("is_windows,expected", [
    (False, "/opt/cycle/jetpack/system/embedded/bin/azcopy"),
    (True, "C:\\cycle\\jetpack\\system\\embedded\\bin\\azcopy.exe"),
])
def test_azcopy(is_windows: bool, expected: str):
    with mock.patch('jetpack.util.is_windows', return_value=is_windows):
        path = azcopy._azcopy()
        assert path == expected


@pytest.mark.parametrize('src_path,dst_path,has_src_locker,has_dst_locker,src_expected,dst_expected', [
    ('/local/file.zip', 'remote/file.zip', False, True, '/local/file.zip', 'https://path/to/remote/file.zip'),
    ('remote/file.tgz', '/local/file.tgz', True, False, 'https://path/to/remote/file.tgz', '/local/file.tgz')
])
def test_get_src_and_dst(src_path: str, dst_path: str, has_src_locker: bool, has_dst_locker: bool, src_expected: str, dst_expected: str):
    src_locker: Optional[Locker] = None
    dst_locker: Optional[Locker] = None
    if has_src_locker:
        src_locker = mock.Mock(spec=Locker)
        src_locker.url.return_value = 'https://path/to/' + src_path
    if has_dst_locker:
        dst_locker = mock.Mock(spec=Locker)
        dst_locker.url.return_value = 'https://path/to/' + dst_path

    src, dst = azcopy._get_src_and_dst(src_path, dst_path, src_locker, dst_locker)
    assert src_expected == src
    assert dst_expected == dst
    if has_src_locker:
        src_locker.url.assert_called_once_with(src_path, include_sas_token=True)  # type: ignore
    if has_dst_locker:
        dst_locker.url.assert_called_once_with(dst_path, include_sas_token=True)  # type: ignore


def test_copy():
    locker = mock.Mock(spec=Locker)
    locker.url.return_value = "https://path/to/remote/file.tgz"
    locker.client_id = "123"

    with mock.patch('jetpack.util.azcopy._azcopy', return_value="azcopy"):
        with mock.patch('jetpack.util.azcopy._execute') as execute:
            azcopy.copy('src/path', 'dst/path', src_locker=locker)
            execute.assert_called_once_with(['copy', 'https://path/to/remote/file.tgz', 'dst/path'], client_id="123")

            execute.reset_mock()
            azcopy.copy('src/path', 'dst/path', dst_locker=locker, recursive=True)
            execute.assert_called_once_with(['copy', 'src/path', 'https://path/to/remote/file.tgz', '--recursive'], client_id="123")


def test_sync():
    locker = mock.Mock(spec=Locker)
    locker.url.return_value = "https://path/to/remote/file.tgz"
    locker.client_id = "123"

    with mock.patch('jetpack.util.azcopy._azcopy', return_value="azcopy"):
        with mock.patch('jetpack.util.azcopy._execute') as execute:
            azcopy.sync('src/path', 'dst/path', src_locker=locker)
            execute.assert_called_once_with(['sync', 'https://path/to/remote/file.tgz', 'dst/path'], client_id="123")

            execute.reset_mock()
            azcopy.sync('src/path', 'dst/path', dst_locker=locker, recursive=True)
            execute.assert_called_once_with(['sync', 'src/path', 'https://path/to/remote/file.tgz', '--recursive'], client_id="123")

            locker.client_id = None
            execute.reset_mock()
            azcopy.sync('src/path', 'dst/path', dst_locker=locker, delete_destination=True)
            execute.assert_called_once_with(['sync', 'src/path', 'https://path/to/remote/file.tgz', '--delete-destination=true'], client_id=None)

def test_get_client_id():
    # no locker, or a locker with no client id
    assert azcopy.get_client_id(None, None) is None
    assert azcopy.get_client_id(make_locker(), None) is None
    assert azcopy.get_client_id(None, make_locker()) is None
    assert azcopy.get_client_id(make_locker(), make_locker()) is None

    # one client id needed
    assert azcopy.get_client_id(make_locker("abc"), None) is "abc"
    assert azcopy.get_client_id(None, make_locker("abc")) is "abc"

    # two client ids, but they match
    assert azcopy.get_client_id(make_locker("abc"), make_locker("abc")) is "abc"

    with pytest.raises(ValueError):
        # client ids do not match
        azcopy.get_client_id(make_locker("abc"), make_locker("def"))

    # override takes precedence even in the case of mismatches
    assert azcopy.get_client_id(make_locker("abc"), make_locker("def"), "ghi") is "ghi"

def make_locker(client_id: Optional[str] = None) -> Locker:
    locker = mock.Mock(spec=Locker)
    locker.client_id = client_id
    return locker
