from subprocess import CalledProcessError
import unittest
from datetime import datetime
from jetpack import util
from jetpack.util import UserError
from unittest import mock
import io
import json
import socket
import ssl


class TestUtil(unittest.TestCase):
    def setUp(self):
        pass

    def test_get_provider(self):
        self.assertEqual("azure", util.get_provider())

    @mock.patch('jetpack.util.system.install_package')
    @mock.patch('jetpack.util.system.execute')
    @mock.patch('jetpack.util.get_selinux_enforce', return_value=True)
    def test_selinux_mode(self, get_selinux_enforce, execute, install_package):

        util.selinux_mode(mode='disabled')
        install_package.assert_called_once_with('libselinux-utils')
        execute.assert_has_calls([
            mock.call("setenforce 0", shell=True),
            mock.call("sed -i 's/^SELINUX=.*/SELINUX=disabled/g' /etc/selinux/config", shell=True)
        ])

        # If selinux isn't currently "enforcing", no need to change the mode
        install_package.reset_mock()
        execute.reset_mock()
        get_selinux_enforce.return_value = False
        util.selinux_mode(mode='disabled')
        self.assertFalse(install_package.called)
        self.assertFalse(execute.called)

        # An invalid mode will just raise a UserError
        with self.assertRaises(UserError) as e:
            get_selinux_enforce.return_value = True
            util.selinux_mode(mode='not a real mode')

    @mock.patch('os.path.exists')
    def test_get_selinux_enforce(self, mock_exists):
        # SELinux is explicitly enabled
        mock_exists.return_value = True
        with mock.patch('jetpack.util.open', mock.mock_open(), create=True) as mock_open:
            mock_open.return_value.__enter__.return_value.read.return_value = '1'
            self.assertTrue(util.get_selinux_enforce())

        # SELinux is excplicity disabled
        mock_exists.return_value = True
        with mock.patch('jetpack.util.open', mock.mock_open(), create=True) as mock_open:
            mock_open.return_value.__enter__.return_value.read.return_value = '0'
            self.assertFalse(util.get_selinux_enforce())
        
        # No files found to explicitly determine selinux status, so assume disabled
        mock_exists.return_value = False
        with mock.patch('jetpack.util.open', mock.mock_open(), create=True) as mock_open:
            mock_open.return_value.__enter__.return_value.read.return_value = '1'
            self.assertFalse(util.get_selinux_enforce())

    def test_write_timestamps(self):
        with mock.patch('os.path.isfile', return_value=False) as mock_isfile:
            
            # First time running, there is no existing timestamp file on disk we should just write what we are given
            with mock.patch('jetpack.util.open', mock.mock_open(), create=True) as mock_open:
                timestamps = util.write_timestamps({'timestamp': datetime(2015, 12, 25)})
                mock_isfile.assert_called_once_with('/opt/cycle/jetpack/config/timestamps.json')
                file_handle = mock_open.return_value.__enter__.return_value
                file_handle.write.assert_called_once_with ('{"timestamp": "2015-12-25T00:00:00"}')
                self.assertEqual({'timestamp': '2015-12-25T00:00:00'}, timestamps)
            
            # Run again, we should merge with the file on disk retaining the original timestamps on file
            mock_isfile.return_value = True
            with mock.patch('jetpack.util.open', mock.mock_open(), create=True) as mock_open:
                mock_open.return_value.__enter__.return_value.read.return_value = """
                {"timestamp": "oldtimedata", "othertimestamp": "data"}
                """
                util.write_timestamps({'timestamp': datetime(2015, 12, 25)})
                self.assertEqual(2, mock_open.call_count)
                file_handle = mock_open.return_value.__enter__.return_value
                file_handle.write.assert_called_once_with ('{"othertimestamp": "data", "timestamp": "oldtimedata"}')

    def test_config_to_dict(self):

        # Build up a config parser and make sure it's converted to a dict
        # correctly.
        import configparser
        parser = configparser.RawConfigParser()
        parser.add_section('states')
        parser.add_section('constants')
        parser.set('states', 'massachusettes', 'boston')
        parser.set('states', 'connecticut', 'hartford')
        parser.set('constants', 'pi', '3.141')
        parser.set('constants', 'e', '2.718')

        result = util._config_to_dict(parser)
        self.assertEqual({
            'states': {
                'massachusettes': 'boston',
                'connecticut': 'hartford'
            },
            "constants": {
                "pi": '3.141',
                "e": '2.718'
            }
        }, result)

    def test_get_config_file(self):
        self.assertEqual("/tmp/test.ini", util._get_config_file("/tmp/test.ini"))
        with mock.patch('platform.system', mock.MagicMock(return_value="Linux")):
            self.assertEqual("/opt/cycle/jetpack/config/jetpack.ini", util._get_config_file())
        with mock.patch('platform.system', mock.MagicMock(return_value="Windows")):
            self.assertEqual("C:\\cycle\\jetpack\\config\\jetpack.ini", util._get_config_file())

    def test_parse_config(self):

        # Verify a missing config yields an empty dict
        with mock.patch('os.path.isfile') as mock_isfile:
            mock_isfile.return_value = False
            self.assertEqual({}, util.parse_config("/tmp/test.ini"))
        # TODO: rewrite this to work with real .ini file
            # mock_isfile.reset_mock()
            # mock_isfile.return_value = True
            # with mock.patch('platform.system', mock.MagicMock(return_value="Windows")):
            #     with mock.patch('jetpack.util._config_to_dict') as config_to_dict:
            #         config_to_dict.return_value = {'a': 1}
            #         self.assertEqual({'a': 1}, util.parse_config())
            #         mock_isfile.assert_called_once_with('C:\\cycle\\jetpack\\config\\jetpack.ini')

            # mock_isfile.reset_mock()
            # mock_isfile.return_value = True
            # with mock.patch('platform.system', mock.MagicMock(return_value="Linux")):
            #     with mock.patch('jetpack.util._config_to_dict') as config_to_dict:
            #         config_to_dict.return_value = {'a': 1}
            #         self.assertEqual({'a': 1}, util.parse_config())
            #         mock_isfile.assert_called_once_with('/opt/cycle/jetpack/config/jetpack.ini')

    def test_write_config(self):
        with mock.patch('jetpack.util._transform_dict_to_ini', return_value="sample ini"):
            with mock.patch('jetpack.util.os.path.exists', return_value=True) as mock_exists:
                with mock.patch('jetpack.util.open', mock.mock_open(), create=True) as mock_open:
                    with mock.patch('os.chmod'):
                        util.write_config({'test': 'dict'}, '/tmp/jetpack.ini')
                        mock_exists.assert_called_once_with('/tmp')
                        mock_open.assert_called_once_with("/tmp/jetpack.ini", "w")
                        mock_open.return_value.write.assert_called_once_with("sample ini")

        with mock.patch('jetpack.util._transform_dict_to_ini', return_value="sample ini"):
            with mock.patch('jetpack.util.os.path.exists', return_value=False) as mock_exists:
                with mock.patch('jetpack.util.open', mock.mock_open(), create=True) as mock_open:
                    with mock.patch('jetpack.util.os.makedirs') as mock_makedirs:
                        with mock.patch('os.chmod'):
                            util.write_config({'test': 'dict'}, '/does/not/exist/jetpack.ini')
                            mock_exists.assert_called_once_with('/does/not/exist')
                            mock_makedirs.assert_called_once_with('/does/not/exist')
                            mock_open.assert_called_once_with("/does/not/exist/jetpack.ini", "w")
                            mock_open.return_value.write.assert_called_once_with("sample ini")

    def test_transform_dict_to_ini(self):
        ini = util._transform_dict_to_ini({
            'cyclecloud': {
                'username': 'my-user', # [SuppressMessage("Microsoft.Security", "CS001:SecretInline", Justification="fake test cred")]
                'password': 'homer simpson'
            },
            'amqp': {
                'port': 8080,
                'hostname': 'localhost'
            }
        })

        expected_ini = """[cyclecloud]
username=my-user # [SuppressMessage("Microsoft.Security", "CS001:SecretInline", Justification="fake test cred")]
password=homer simpson

[amqp]
port=8080
hostname=localhost"""

        # take out the inline comment so we have the actual expected ini...
        expected_ini = "\n".join([l.split(" # ")[0] for l in expected_ini.split("\n")])

        self.assertEqual(ini, expected_ini)

    def test_pid_running(self):
        # Linux
        with mock.patch('platform.system', mock.MagicMock(return_value="Linux")):
            with mock.patch('jetpack.util.os.kill') as mock_kill:
                # Test running
                self.assertTrue(util.pid_running(1337))
                mock_kill.assert_called_once_with(1337, 0)

                mock_kill.reset_mock()
                mock_kill.side_effect = OSError("PID doesn't exist!")
                self.assertFalse(util.pid_running(1337))
                mock_kill.assert_called_once_with(1337, 0)

        # windows
        with mock.patch('platform.system', mock.MagicMock(return_value="Windows")):
            with mock.patch('jetpack.util.ctypes') as mock_ctypes:
                kernel32 = mock.Mock()
                mock_ctypes.windll.kernel32 = kernel32
                kernel32.OpenProcess.return_value = 1  # Running
                self.assertTrue(util.pid_running(12345))
                kernel32.OpenProcess.assert_called_once_with (0x100000, 0, 12345)
                kernel32.CloseHandle.assert_called_once_with (1)

                kernel32.reset_mock()
                mock_ctypes.reset_mock()
                mock_ctypes.windll.kernel32 = kernel32
                kernel32.OpenProcess.return_value = 0  # Not Running
                kernel32.OpenProcess.assert_not_called()
                self.assertFalse(util.pid_running(12345))

    @mock.patch('jetpack.config.get', return_value=False)
    @mock.patch('jetpack.util.get_connection_config')
    def test_get_cyclecloud_connection(self, get_connection_config, mock_get):
        get_connection_config.return_value = {
            'url': 'http://127.0.0.1:8080',
            'username': 'test',
            'password': 'password'            
        }
                
        conn, headers = util.get_cyclecloud_connection()
        self.assertEqual("127.0.0.1", conn.host)
        self.assertEqual(8080, conn.port)
        # [SuppressMessage("Microsoft.Security", "CS002:SecretInNextLine", Justification="fake test cred")]
        self.assertEqual({'Authorization': 'Basic dGVzdDpwYXNzd29yZA==', 'User-Agent': 'azure/jetpack'}, headers)

        # # HTTPS with ssl validation
        get_connection_config.return_value = {
            'url': 'https://127.0.0.1:8443',
            'username': 'test',
            'password': 'password',
            'cert': 'test-cert'
        }
        with mock.patch('ssl.create_default_context') as create_default_context:
            conn, headers = util.get_cyclecloud_connection()
            self.assertEqual("127.0.0.1", conn.host)
            self.assertEqual(8443, conn.port)
            create_default_context.assert_called_once_with(ssl.Purpose.SERVER_AUTH, cadata='test-cert')

        # HTTPS without ssl validation
        get_connection_config.return_value = {
            'url': 'https://127.0.0.1:8443',
            'username': 'test',
            'password': 'password'
        }

        conn, headers = util.get_cyclecloud_connection()
        self.assertEqual(util._unverified_context, conn._context)

    def test_get_boottime(self):
        DECEMBER_24_2015 = 1450933200
        DECEMBER_25_2015 = 1451019600
        DAY_IN_SECS = 86400
        with mock.patch('platform.system', mock.MagicMock(return_value="Linux")) as mock_system:
            with mock.patch('jetpack.util.open', mock.mock_open(), create=True) as mock_open:
                with mock.patch('jetpack.util.time.time', return_value=DECEMBER_25_2015) as mock_time:
                    mock_open.return_value.readline.return_value = "%s" % DAY_IN_SECS * 1
                    self.assertEqual(datetime.fromtimestamp(DECEMBER_24_2015), util.get_boottime())
                    mock_open.assert_any_call("/proc/uptime", 'r')

        with mock.patch('platform.system', mock.MagicMock(return_value="Windows")) as mock_system:
            with mock.patch('jetpack.util.time.time', return_value=DECEMBER_25_2015) as mock_time:
                with mock.patch('jetpack.util.ctypes') as mock_ctypes: 
                    kernel32 = mock.Mock()
                    mock_ctypes.windll.kernel32 = kernel32

                    # Note: Windows GetTickCount64 is in milliseconds
                    kernel32.GetTickCount64.return_value = float(DAY_IN_SECS * 1 * 1000)

                    self.assertEqual(datetime.fromtimestamp(DECEMBER_24_2015), util.get_boottime()) 

    @mock.patch("jetpack.util.os.path.isfile")
    def test_stale_file_new(self, mock_isfile):

        with mock.patch('jetpack.util.get_boottime') as mock_boottime:
            with mock.patch('jetpack.util.os.path.getmtime') as mock_getmtime:
                
                # The file doesn't exist, means it's not stale
                mock_isfile.return_value = False
                self.assertFalse(util.stale_file(123))

                # boottime is newer than mtime means the file is stale
                mock_isfile.return_value = True
                mock_boottime.return_value = datetime(2015, 12, 25)
                mock_getmtime.return_value = int(datetime(2015, 12, 24).strftime("%s"))
                self.assertTrue(util.stale_file(123))

                # boottime not newer than mtime means the file is not stale
                mock_isfile = True
                mock_boottime.return_value = datetime(2015, 12, 25)
                mock_getmtime.return_value = int(datetime(2015, 12, 26).strftime("%s"))
                self.assertFalse(util.stale_file(123))

    def test_already_running(self):

        # 1: Pid file doesn't exist at all
        with mock.patch('jetpack.util.os.path.isfile', return_value=False) as mock_isfile:
            with mock.patch('jetpack.util.open', mock.mock_open(), create=True) as mock_open:
                with mock.patch('jetpack.util.os.getpid', return_value=1337) as mock_getpid:
                    self.assertFalse(util.already_running("/path/to/test.pid"))
                    mock_isfile.assert_called_once_with("/path/to/test.pid")
                    mock_open.assert_called_once_with('/path/to/test.pid', 'w')
                    mock_getpid.assert_called_once()
                    mock_open.return_value.write.assert_called_once_with("1337")

        # 2: Pid file exists and isn't running
        with mock.patch('jetpack.util.os.path.isfile', return_value=True) as mock_isfile:
            with mock.patch('jetpack.util.open', mock.mock_open(), create=True) as mock_open:
                with mock.patch('jetpack.util.os.getpid', return_value=1337) as mock_getpid:
                    with mock.patch('jetpack.util.pid_running', return_value=False) as mock_pid_running:
                        with mock.patch('jetpack.util.stale_file', return_value=False) as mock_stale_file:
                            mock_open.return_value.read.return_value = "50"
                            self.assertFalse(util.already_running("/path/to/test.pid"))
                            mock_isfile.assert_called_once_with("/path/to/test.pid")
                            mock_open.assert_any_call("/path/to/test.pid")
                            mock_open.return_value.read.assert_called_once()
                            mock_pid_running.assert_called_once_with(50)
                            mock_open.assert_any_call("/path/to/test.pid", "w")
                            mock_getpid.assert_called_once()
                            mock_open.return_value.write.assert_called_once_with("1337")
                            mock_stale_file.assert_called_once_with("/path/to/test.pid")

        # 3: Pid file exists and IS running
        with mock.patch('jetpack.util.os.path.isfile', return_value=True) as mock_isfile:
            with mock.patch('jetpack.util.open', mock.mock_open(), create=True) as mock_open:
                with mock.patch('jetpack.util.pid_running', return_value=True) as mock_pid_running:
                    with mock.patch('jetpack.util.stale_file', return_value=False) as mock_stale_file:
                        mock_open.return_value.read.return_value = "50"
                        self.assertTrue(util.already_running("/path/to/test.pid"))
                        mock_isfile.assert_called_once_with("/path/to/test.pid")
                        mock_open.assert_called_once_with("/path/to/test.pid")
                        mock_open.return_value.read.assert_called_once()
                        mock_pid_running.assert_called_once_with(50)
                        mock_stale_file.assert_called_once_with("/path/to/test.pid")
                        
        # 4. Pid file exists and is running but is stale
        with mock.patch('jetpack.util.os.path.isfile', return_value=True) as mock_isfile:
            with mock.patch('jetpack.util.pid_running', return_value=True) as mock_pid_running:
                with mock.patch('jetpack.util.stale_file', return_value=True) as mock_stale_file:
                    with mock.patch('jetpack.util.open', mock.mock_open(), create=True) as mock_open:
                        self.assertFalse(util.already_running("/path/to/test.pid"))

        # Verifies the fix for Gitlab Issue #66
        # 5. PID file exists and is empty / not an integer
        with mock.patch('jetpack.util.os.path.isfile', return_value=True) as mock_isfile:
            with mock.patch('jetpack.util.stale_file', return_value=False) as mock_stale_file:
                with mock.patch('jetpack.util.open', mock.mock_open(), create=True) as mock_open:
                    mock_open.return_value.read.return_value = ""                    
                    self.assertFalse(util.already_running("/path/to/test.pid"))
                    mock_open.return_value.read.return_value = "not an integer"
                    self.assertFalse(util.already_running("/path/to/test.pid"))

    def test_deep_merge(self):
        src = {
            "nested": {
                "a": 1, 
                "b": 2
            },
            "test": True
        }

        dst = {
            "nested": {
                "b": 3,
                "c": 4
            },
            "test-2": ["pass"]
        }

        util.deep_merge(src, dst)
        self.assertEqual({
            "nested": {
                "a": 1,
                "b": 2,
                "c": 4
            },
            "test": True,
            "test-2": ["pass"]
        }, dst)

    @mock.patch('os.chown')
    @mock.patch('os.chmod')
    @mock.patch('os.stat')
    def test_update_file_permissions(self, mock_stat, mock_chmod, mock_chown):
        mock_stat.return_value.st_uid = 123
        mock_stat.return_value.st_gid = 456
        mock_stat.return_value.st_mode = 0o100755
        util.update_file_permissions('/tmp/test.file', 123, 456, 0o755)
        mock_chmod.assert_not_called()
        mock_chown.assert_not_called()
        mock_chmod.reset_mock()
        mock_chown.reset_mock()
        mock_stat.return_value.st_uid = 124
        mock_stat.return_value.st_gid = 452
        mock_stat.return_value.st_mode = 0o100754
        util.update_file_permissions('/tmp/test.file', 123, 456, 0o755)
        mock_chmod.assert_called_once_with('/tmp/test.file', 0o755)
        mock_chown.assert_called_once_with('/tmp/test.file', 123, 456)

    def test_separate(self):
        self.assertEqual(['/', 'this', 'is', 'a', 'path'], util.separate('/this/is/a/path'))
        self.assertEqual(['/'], util.separate('/'))
        self.assertEqual([''], util.separate(''))

    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('jetpack.util.jetpack_home', return_value='/tmp/jetpack')
    def test_version(self, mock_jetpack_home, mock_exists):
        with mock.patch('jetpack.util.open', mock.mock_open(), create=True) as mock_open:
            mock_open.return_value.read.return_value = '1.2.3'
            self.assertEqual('1.2.3', util.version())
            mock_exists.assert_called_once_with('/tmp/jetpack/version.txt')
            mock_open.assert_called_once_with('/tmp/jetpack/version.txt', 'r')

        mock_exists.return_value = False
        with self.assertRaises(Exception):
            util.version()

    def test_commonpath(self):
        self.assertEqual('/', util.commonpath(['/foo', '/bar']))
        self.assertEqual('/test', util.commonpath(['/test/1/2/3', '/test/4']))
        self.assertEqual('/', util.commonpath(['/test', '/testing']))
        self.assertEqual('/test/ing', util.commonpath(['/test/ing/1', '/test/ing/1/2', '/test/ing']))

    def test_query_compute_metadata_success(self):
        with mock.patch('urllib.request.urlopen') as mock_open:
            mock_open.return_value.getcode.return_value = 200
            mock_open.return_value.read.return_value = json.dumps({'testing': 123})
            self.assertEqual({'testing': 123}, util._query_compute_metadata())

    @mock.patch('time.sleep')
    def test_query_compute_metadata_failure(self, mock_sleep):
        # The _query_compute_metadata() method has a retry decorator on it
        # to try 5 times, make sure the decorator's sleep method is called 4 times
        # i.e. 5 tries
        with mock.patch('urllib.request.urlopen', side_effect=Exception("oh noes")):
            with self.assertRaises(Exception):
                util._query_compute_metadata()
            self.assertEqual(4, mock_sleep.call_count)

    def test_get_instance_id(self):
        with mock.patch('jetpack.util._query_compute_metadata') as mock_metadata:
            # regular VM
            mock_metadata.return_value = json.loads('''{"location":"southcentralus","name":"some-name-here","offer":"CentOS-HPC","osType":"Linux",
                "resourceGroupName":"custom-resource-group","sku":"7.6","subscriptionId":"d7e99a8b-42da-4270-bb84-e85bdaa4491e","tags":"",
                "version":"7.6.20190721","vmId":"","vmSize":"Standard_D8s_v3"}''')
            self.assertEqual("1d836d60cbe91db00fbc79a890324628", util.get_instance_id())

            # VMSS
            mock_metadata.return_value = json.loads('''{"location":"southcentralus","name":"some-name-here_1","offer":"CentOS-HPC","osType":"Linux",
                "resourceGroupName":"custom-resource-group","sku":"7.6","subscriptionId":"d7e99a8b-42da-4270-bb84-e85bdaa4491e","tags":"",
                "version":"7.6.20190721","vmId":"","vmSize":"Standard_D8s_v3"}''')
            self.assertEqual("50de3f29506b00801584f536afefc7b9", util.get_instance_id())

    def test_get_resource_name(self):
        with mock.patch('jetpack.util._query_compute_metadata') as mock_metadata:
            # regular VM
            mock_metadata.return_value = json.loads('''{"location":"southcentralus","name":"some-name-here","offer":"CentOS-HPC","osType":"Linux",
                "resourceGroupName":"custom-resource-group","sku":"7.6","subscriptionId":"d7e99a8b-42da-4270-bb84-e85bdaa4491e","tags":"",
                "version":"7.6.20190721","vmId":"","vmSize":"Standard_D8s_v3"}''')
            self.assertEqual("some-name-here", util.get_resource_name())

            # VMSS
            mock_metadata.return_value = json.loads('''{"location":"southcentralus","name":"some-name-here_1","offer":"CentOS-HPC","osType":"Linux",
                "resourceGroupName":"custom-resource-group","sku":"7.6","subscriptionId":"d7e99a8b-42da-4270-bb84-e85bdaa4491e","tags":"",
                "version":"7.6.20190721","vmId":"","vmSize":"Standard_D8s_v3"}''')
            self.assertEqual("some-name-here_1", util.get_resource_name())

    @mock.patch('jetpack.util.get_connection_config')
    @mock.patch('jetpack.util.TOTAL_ATTEMPT_TIME_THROUGH_PROXY', .1)
    @mock.patch('jetpack.util.TOTAL_ATTEMPT_TIME', .1)
    @mock.patch('jetpack.util.SOCKET_TIMEOUT', .1)
    @mock.patch('jetpack.util._connect_to_cyclecloud')
    @mock.patch('jetpack.util.get_instance_id')
    def test_test_cyclecloud_connection(self, mock_instance_id, mock_connect, mock_get_connection_config):
        connection = {"cluster": "c", "url": "http://servername:1000", "cert": "mycert"}

        connection["proxy_method"] = "none"

        mock_get_connection_config.return_value = connection
        mock_instance_id.return_value = "i-123"
        mock_connect.side_effect = socket.timeout('Error')
        with self.assertRaises(UserError) as error:
            util.test_cyclecloud_connection()

        self.assertEqual("Timed out connecting to CycleCloud at http://servername:1000", str(error.exception))
        self.assertEqual(None, error.exception.recommendation)

        connection["proxy_method"] = "local"
        mock_get_connection_config.return_value = connection
        with self.assertRaises(UserError) as error:
            util.test_cyclecloud_connection()
        self.assertEqual("Timed out connecting to CycleCloud through return-proxy tunnel at http://servername:1000", str(error.exception))
        self.assertEqual("Make sure CycleCloud has established a return-proxy tunnel.", error.exception.recommendation)

        connection["proxy_method"] = "none"
        mock_get_connection_config.return_value = connection
        mock_connect.side_effect = socket.error('Some networking error')
        with self.assertRaises(UserError) as error:
            util.test_cyclecloud_connection()
        self.assertEqual("Could not connect to CycleCloud at http://servername:1000: Some networking error", str(error.exception))

        connection["proxy_method"] = "remote"
        mock_get_connection_config.return_value = connection
        with self.assertRaises(UserError) as error:
            util.test_cyclecloud_connection()
        self.assertEqual("Could not connect to CycleCloud through return-proxy tunnel at http://servername:1000: Some networking error", str(error.exception))

        connection["proxy_method"] = "none"
        mock_get_connection_config.return_value = connection
        mock_connect.side_effect.errno = 111
        with self.assertRaises(UserError) as error:
            util.test_cyclecloud_connection()
        self.assertEqual("Connection refused to CycleCloud at http://servername:1000", str(error.exception))

        mock_connect.side_effect = ssl.SSLError('This failed')
        mock_connect.side_effect.reason = "CERTIFICATE_VERIFY_FAILED"
        with self.assertRaises(UserError) as error:
            util.test_cyclecloud_connection()
        self.assertEqual("Certificate validation failed for CycleCloud at http://servername:1000", str(error.exception))

        # unexpected status
        mock_connect.side_effect = None
        mock_connect.return_value.status = 500
        mock_connect.return_value.read.return_value = '{}'
        with self.assertRaises(UserError) as error:
            util.test_cyclecloud_connection()
        self.assertEqual("Unexpected error connecting to CycleCloud at http://servername:1000: status code 500", str(error.exception))

        # success, no exception
        mock_connect.side_effect = None
        mock_connect.return_value.status = 200
        mock_connect.return_value.read.return_value = '{}'
        util.test_cyclecloud_connection()

        # cluster name / instance_id with spaces should be properly escaped when making a connection
        connection['cluster'] = 'name with spaces'
        mock_get_connection_config.return_value = connection
        mock_instance_id.return_value = "i 123"
        mock_connect.reset_mock()
        mock_connect.side_effect = None
        mock_connect.return_value.status = 200
        mock_connect.return_value.read.return_value = '{}'
        util.test_cyclecloud_connection()
        mock_connect.assert_called_once_with('/cloud/api/node_status/i%20123?cluster=name%20with%20spaces')

    @mock.patch('subprocess.check_call')
    @mock.patch('subprocess.check_output')
    @mock.patch('os.path.exists')
    def test_rebuild_rpmdb(self, exists, check_output, check_call):

        # Note RHEL - don't even check
        exists.return_value = False
        self.assertFalse(util.rebuild_rpmdb())
        exists.assert_called_once_with("/etc/redhat-release")
        self.assertFalse(check_output.called)
        self.assertFalse(check_call.called)

        # RHEL and rpmdb isn't corrupt
        exists.reset_mock()
        check_call.reset_mock()
        check_output.reset_mock()

        exists.return_value = True
        check_output.return_value = "Output from rpm -qa"

        self.assertFalse(util.rebuild_rpmdb())
        self.assertFalse(check_call.called)

        # RHEL and rpmdb is corrupt, so it should be rebuilt
        exists.reset_mock()
        check_call.reset_mock()
        check_output.reset_mock()

        exists.return_value = True
        check_output.side_effect = CalledProcessError(1, "some command", "The database was corrupt, run database recovery")
        self.assertTrue(util.rebuild_rpmdb())
        check_call.assert_called_once_with(['rpm', '--rebuilddb'])

    @mock.patch('jetpack.util.get_connection_config', return_value={'url': 'http://localhost:9443/'})
    @mock.patch('jetpack.util.get_cyclecloud_connection')
    def test_query_cyclecloud(self, mock_get_cyclecloud_connection, mock_get_connection_config):
        conn = mock.Mock()
        mock_get_cyclecloud_connection.return_value = (conn, {'Authorization': 'Nope!'})

        # Simple GET
        util.query_cyclecloud("/query/path", headers={'additional': 'header'}, params={"awesome": "yes"}, method="GET")
        conn.request.assert_called_once_with("GET", "/query/path?awesome=yes", body=None, headers={"Authorization": "Nope!", "additional": "header"})

        # Simple POST w/ body
        conn.reset_mock()
        mock_get_cyclecloud_connection.return_value = (conn, {'Authorization': 'Nope!'})
        util.query_cyclecloud("/post/path", body="CONTENT", headers={'Authorization': 'overwritten'}, method="POST")
        conn.request.assert_called_once_with("POST", "/post/path", body="CONTENT", headers={'Authorization': 'overwritten'})

        # w/ Context path
        conn.reset_mock()
        mock_get_cyclecloud_connection.return_value = (conn, {'Authorization': 'Nope!'})
        mock_get_connection_config.return_value = {'url': 'https://localhost:9443/cyclecloud'}
        util.query_cyclecloud("/some/path")
        conn.request.assert_called_once_with("GET", "/cyclecloud/some/path", body=None, headers={'Authorization': 'Nope!'})

    def test_get_nodearray(self):
         with mock.patch('jetpack.util._query_compute_metadata') as mock_metadata:
            # regular VM
            mock_metadata.return_value = json.loads('''{"location":"southcentralus","name":"scheduler-1d836d60cbe91db00fbc79a8","offer":"CentOS-HPC","osType":"Linux",
                "resourceGroupName":"custom-resource-group","sku":"7.6","subscriptionId":"d7e99a8b-42da-4270-bb84-e85bdaa4491e","tags":"",
                "version":"7.6.20190721","vmId":"","vmSize":"Standard_D8s_v3"}''')
            self.assertEqual("scheduler", util.get_nodearray_name())

            # with a dash in the name
            mock_metadata.return_value = json.loads('''{"location":"southcentralus","name":"primary-scheduler-1d836d60cbe91db00fbc79a8","offer":"CentOS-HPC","osType":"Linux",
                "resourceGroupName":"custom-resource-group","sku":"7.6","subscriptionId":"d7e99a8b-42da-4270-bb84-e85bdaa4491e","tags":"",
                "version":"7.6.20190721","vmId":"","vmSize":"Standard_D8s_v3"}''')
            self.assertEqual("primary-scheduler", util.get_nodearray_name())

            # VMSS
            mock_metadata.return_value = json.loads('''{"location":"southcentralus","name":"hpc-jn123kj1l23n412asnbq_1","offer":"CentOS-HPC","osType":"Linux",
                "resourceGroupName":"custom-resource-group","sku":"7.6","subscriptionId":"d7e99a8b-42da-4270-bb84-e85bdaa4491e","tags":"",
                "version":"7.6.20190721","vmId":"","vmSize":"Standard_D8s_v3"}''')
            self.assertEqual("hpc", util.get_nodearray_name())
